# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Computes a Best-Response policy.

The goal if this file is to be the main entry-point for BR APIs in Python.

TODO(author2): Also include computation using the more efficient C++
`TabularBestResponse` implementation.
"""

import collections
import itertools

import numpy as np

from open_spiel.python import games  # pylint:disable=unused-import
from open_spiel.python import policy as openspiel_policy
from open_spiel.python.algorithms import get_all_states
from open_spiel.python.algorithms import noisy_policy
from open_spiel.python.algorithms import policy_utils
import pyspiel


def _memoize_method(key_fn=lambda x: x):
  """Memoize a single-arg instance method using an on-object cache."""

  def memoizer(method):
    cache_name = "cache_" + method.__name__

    def wrap(self, arg):
      key = key_fn(arg)
      cache = vars(self).setdefault(cache_name, {})
      if key not in cache:
        cache[key] = method(self, arg)
      return cache[key]

    return wrap

  return memoizer


def compute_states_and_info_states_if_none(game,
                                           all_states=None,
                                           state_to_information_state=None):
  """Returns all_states and/or state_to_information_state for the game.

  To recompute everything, pass in None for both all_states and
  state_to_information_state. Otherwise, this function will use the passed in
  values to reconstruct either of them.

  Args:
    game: The open_spiel game.
    all_states: The result of calling get_all_states.get_all_states. Cached for
      improved performance.
    state_to_information_state: A dict mapping state.history_str() to
      state.information_state for every state in the game. Cached for improved
      performance.
  """
  if all_states is None:
    all_states = get_all_states.get_all_states(
        game,
        depth_limit=-1,
        include_terminals=False,
        include_chance_states=False)

  if state_to_information_state is None:
    state_to_information_state = {
        state: all_states[state].information_state_string()
        for state in all_states
    }

  return all_states, state_to_information_state


class BestResponsePolicy(openspiel_policy.Policy):
  """Computes the best response to a specified strategy."""

  def __init__(self,
               game,
               player_id,
               policy,
               root_state=None,
               cut_threshold=0.0):
    """Initializes the best-response calculation.

    Args:
      game: The game to analyze.
      player_id: The player id of the best-responder.
      policy: A `policy.Policy` object.
      root_state: The state of the game at which to start analysis. If `None`,
        the game root state is used.
      cut_threshold: The probability to cut when calculating the value.
        Increasing this value will trade off accuracy for speed.
    """
    self._num_players = game.num_players()
    self._player_id = player_id
    self._policy = policy
    if root_state is None:
      root_state = game.new_initial_state()
    self._root_state = root_state
    self.infosets = self.info_sets(root_state)

    self._cut_threshold = cut_threshold

  def info_sets(self, state):
    """Returns a dict of infostatekey to list of (state, cf_probability)."""
    infosets = collections.defaultdict(list)
    for s, p in self.decision_nodes(state):
      infosets[s.information_state_string(self._player_id)].append((s, p))
    return dict(infosets)

  def decision_nodes(self, parent_state):
    """Yields a (state, cf_prob) pair for each descendant decision node."""
    if not parent_state.is_terminal():
      if (parent_state.current_player() == self._player_id or
          parent_state.is_simultaneous_node()):
        yield (parent_state, 1.0)
      for action, p_action in self.transitions(parent_state):
        for state, p_state in self.decision_nodes(
            openspiel_policy.child(parent_state, action)):
          yield (state, p_state * p_action)

  def joint_action_probabilities_counterfactual(self, state):
    """Get list of action, probability tuples for simultaneous node.

    Counterfactual reach probabilities exclude the best-responder's actions,
    the sum of the probabilities is equal to the number of actions of the
    player _player_id.
    Args:
      state: the current state of the game.

    Returns:
      list of action, probability tuples. An action is a tuple of individual
        actions for each player of the game.
    """
    actions_per_player, probs_per_player = (
        openspiel_policy.joint_action_probabilities_aux(state, self._policy))
    probs_per_player[self._player_id] = [
        1.0 for _ in probs_per_player[self._player_id]
    ]
    return [(list(actions), np.prod(probs)) for actions, probs in zip(
        itertools.product(
            *actions_per_player), itertools.product(*probs_per_player))]

  def transitions(self, state):
    """Returns a list of (action, cf_prob) pairs from the specified state."""
    if state.current_player() == self._player_id:
      # Counterfactual reach probabilities exclude the best-responder's actions,
      # hence return probability 1.0 for every action.
      return [(action, 1.0) for action in state.legal_actions()]
    elif state.is_chance_node():
      return state.chance_outcomes()
    elif state.is_simultaneous_node():
      return self.joint_action_probabilities_counterfactual(state)
    else:
      return list(self._policy.action_probabilities(state).items())

  @_memoize_method(key_fn=lambda state: state.history_str())
  def value(self, state):
    """Returns the value of the specified state to the best-responder."""
    if state.is_terminal():
      return state.player_return(self._player_id)
    elif (state.current_player() == self._player_id or
          state.is_simultaneous_node()):
      action = self.best_response_action(
          state.information_state_string(self._player_id))
      return self.q_value(state, action)
    else:
      return sum(p * self.q_value(state, a)
                 for a, p in self.transitions(state)
                 if p > self._cut_threshold)

  def q_value(self, state, action):
    """Returns the value of the (state, action) to the best-responder."""
    if state.is_simultaneous_node():

      def q_value_sim(sim_state, sim_actions):
        child = sim_state.clone()
        # change action of _player_id
        sim_actions[self._player_id] = action
        child.apply_actions(sim_actions)
        return self.value(child)

      actions, probabilities = zip(*self.transitions(state))
      return sum(p * q_value_sim(state, a)
                 for a, p in zip(actions, probabilities / sum(probabilities))
                 if p > self._cut_threshold)
    else:
      return self.value(state.child(action))

  @_memoize_method()
  def best_response_action(self, infostate):
    """Returns the best response for this information state."""
    infoset = self.infosets[infostate]
    # Get actions from the first (state, cf_prob) pair in the infoset list.
    # Return the best action by counterfactual-reach-weighted state-value.
    return max(
        infoset[0][0].legal_actions(self._player_id),
        key=lambda a: sum(cf_p * self.q_value(s, a) for s, cf_p in infoset))

  def action_probabilities(self, state, player_id=None):
    """Returns the policy for a player in a state.

    Args:
      state: A `pyspiel.State` object.
      player_id: Optional, the player id for whom we want an action. Optional
        unless this is a simultaneous state at which multiple players can act.

    Returns:
      A `dict` of `{action: probability}` for the specified player in the
      supplied state.
    """
    if player_id is None:
      if state.is_simultaneous_node():
        player_id = self._player_id
      else:
        player_id = state.current_player()
    return {
        self.best_response_action(state.information_state_string(player_id)): 1
    }


class CPPBestResponsePolicy(openspiel_policy.Policy):
  """Computes best response action_probabilities using open_spiel's C++ backend.

     May have better performance than best_response.py for large games.
  """

  def __init__(self,
               game,
               best_responder_id,
               policy,
               all_states=None,
               state_to_information_state=None,
               best_response_processor=None,
               cut_threshold=0.0):
    """Constructor.

    Args:
      game: The game to analyze.
      best_responder_id: The player id of the best-responder.
      policy: A policy.Policy object representing the joint policy, taking a
        state and returning a list of (action, probability) pairs. This could be
        aggr_policy, for instance.
      all_states: The result of calling get_all_states.get_all_states. Cached
        for improved performance.
      state_to_information_state: A dict mapping state.history_str to
        state.information_state for every state in the game. Cached for improved
        performance.
      best_response_processor: A TabularBestResponse object, used for processing
        the best response actions.
      cut_threshold: The probability to cut when calculating the value.
        Increasing this value will trade off accuracy for speed.
    """
    (self.all_states, self.state_to_information_state) = (
        compute_states_and_info_states_if_none(game, all_states,
                                               state_to_information_state))

    policy_to_dict = policy_utils.policy_to_dict(
        policy, game, self.all_states, self.state_to_information_state)

    # pylint: disable=g-complex-comprehension
    # Cache TabularBestResponse for players, due to their costly construction
    # TODO(b/140426861): Use a single best-responder once the code supports
    # multiple player ids.
    if not best_response_processor:
      best_response_processor = pyspiel.TabularBestResponse(
          game, best_responder_id, policy_to_dict)

    self._policy = policy
    self.game = game
    self.best_responder_id = best_responder_id
    self.tabular_best_response_map = (
        best_response_processor.get_best_response_actions())

    self._cut_threshold = cut_threshold

  def decision_nodes(self, parent_state):
    """Yields a (state, cf_prob) pair for each descendant decision node."""
    if not parent_state.is_terminal():
      if parent_state.current_player() == self.best_responder_id:
        yield (parent_state, 1.0)
      for action, p_action in self.transitions(parent_state):
        for state, p_state in self.decision_nodes(parent_state.child(action)):
          yield (state, p_state * p_action)

  def transitions(self, state):
    """Returns a list of (action, cf_prob) pairs from the specified state."""
    if state.current_player() == self.best_responder_id:
      # Counterfactual reach probabilities exclude the best-responder's actions,
      # hence return probability 1.0 for every action.
      return [(action, 1.0) for action in state.legal_actions()]
    elif state.is_chance_node():
      return state.chance_outcomes()
    else:
      return list(self._policy.action_probabilities(state).items())

  @_memoize_method(key_fn=lambda state: state.history_str())
  def value(self, state):
    """Returns the value of the specified state to the best-responder."""
    if state.is_terminal():
      return state.player_return(self.best_responder_id)
    elif state.current_player() == self.best_responder_id:
      action = self.best_response_action(
          state.information_state_string(self.best_responder_id))
      return self.q_value(state, action)
    else:
      return sum(p * self.q_value(state, a)
                 for a, p in self.transitions(state)
                 if p > self._cut_threshold)

  def q_value(self, state, action):
    """Returns the value of the (state, action) to the best-responder."""
    return self.value(state.child(action))

  @_memoize_method()
  def best_response_action(self, infostate):
    """Returns the best response for this information state."""
    action = self.tabular_best_response_map[infostate]
    return action

  def action_probabilities(self, state, player_id=None):
    """Returns the policy for a player in a state.

    Args:
      state: A `pyspiel.State` object.
      player_id: Optional, the player id for whom we want an action. Optional
        unless this is a simultabeous state at which multiple players can act.

    Returns:
      A `dict` of `{action: probability}` for the specified player in the
      supplied state.
    """
    # Send the best-response probabilities for the best-responder
    if state.current_player() == self.best_responder_id:
      probs = {action_id: 0. for action_id in state.legal_actions()}
      info_state = self.state_to_information_state[state.history_str()]
      probs[self.tabular_best_response_map[info_state]] = 1.
      return probs

    # Send the default probabilities for all other players
    return self._policy.action_probabilities(state, player_id)

  @property
  def policy(self):
    return self._policy

  def copy_with_noise(self, alpha=0.0, beta=0.0):
    """Copies this policy and adds noise, making it a Noisy Best Response.

    The policy's new probabilities P' on each state s become
    P'(s) = alpha * epsilon + (1-alpha) * P(s)

    With P the former policy's probabilities, and epsilon ~ Softmax(beta *
    Uniform)

    Args:
      alpha: First mixture component
      beta: Softmax 1/temperature component

    Returns:
      Noisy copy of best response.
    """
    return noisy_policy.NoisyPolicy(self, alpha, beta, self.all_states)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example running AlphaRank on OpenSpiel games.

  AlphaRank output variable names corresponds to the following paper:
    https://arxiv.org/abs/1903.01373
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app

from open_spiel.python.algorithms import fictitious_play
from open_spiel.python.egt import alpharank
from open_spiel.python.egt import alpharank_visualizer
from open_spiel.python.egt import utils
import pyspiel


def get_kuhn_poker_data(num_players=3):
  """Returns the kuhn poker data for the number of players specified."""
  game = pyspiel.load_game('kuhn_poker', {'players': num_players})
  xfp_solver = fictitious_play.XFPSolver(game, save_oracles=True)
  for _ in range(3):
    xfp_solver.iteration()

  # Results are seed-dependent, so show some interesting cases
  if num_players == 2:
    meta_games = xfp_solver.get_empirical_metagame(100, seed=1)
  elif num_players == 3:
    meta_games = xfp_solver.get_empirical_metagame(100, seed=5)
  elif num_players == 4:
    meta_games = xfp_solver.get_empirical_metagame(100, seed=2)

  # Metagame utility matrices for each player
  payoff_tables = []
  for i in range(num_players):
    payoff_tables.append(meta_games[i])
  return payoff_tables


def main(unused_arg):
  # Construct meta-game payoff tables
  payoff_tables = get_kuhn_poker_data()
  payoffs_are_hpt_format = utils.check_payoffs_are_hpt(payoff_tables)
  strat_labels = utils.get_strat_profile_labels(payoff_tables,
                                                payoffs_are_hpt_format)

  # Run AlphaRank
  rhos, rho_m, pi, _, _ = alpharank.compute(payoff_tables, alpha=1e2)

  # Report & plot results
  alpharank.print_results(
      payoff_tables, payoffs_are_hpt_format, rhos=rhos, rho_m=rho_m, pi=pi)
  utils.print_rankings_table(payoff_tables, pi, strat_labels)
  m_network_plotter = alpharank_visualizer.NetworkPlot(
      payoff_tables, rhos, rho_m, pi, strat_labels, num_top_profiles=8)
  m_network_plotter.compute_and_draw_network()


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fictitious play on an MFG game."""
from typing import Sequence

from absl import flags

from open_spiel.python.mfg.algorithms import fictitious_play
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.games import factory
from open_spiel.python.utils import app
from open_spiel.python.utils import metrics

FLAGS = flags.FLAGS

flags.DEFINE_string('game_name', 'mfg_crowd_modelling_2d', 'Name of the game.')
flags.DEFINE_string(
    'setting', None,
    'Name of the game settings. If None, the game name will be used.')
flags.DEFINE_integer('num_iterations', 100,
                     'Number of fictitious play iterations.')

flags.DEFINE_float('learning_rate', None,
                   'Learning rate. If not, it will be set to 1/iteration.')
flags.DEFINE_string(
    'logdir', None,
    'Logging dir to use for TF summary files. If None, the metrics will only '
    'be logged to stderr.')


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  game = factory.create_game_with_setting(FLAGS.game_name, FLAGS.setting)

  # Metrics writer will also log the metrics to stderr.
  just_logging = FLAGS.logdir is None
  writer = metrics.create_default_writer(
      logdir=FLAGS.logdir, just_logging=just_logging)

  # Save the parameters.
  learning_rate = FLAGS.learning_rate
  writer.write_hparams({'learning_rate': learning_rate})

  fp = fictitious_play.FictitiousPlay(game)

  for it in range(FLAGS.num_iterations):
    fp.iteration(learning_rate=learning_rate)
    fp_policy = fp.get_policy()
    exploitability = nash_conv.NashConv(game, fp_policy).nash_conv()
    writer.write_scalars(it, {'exploitability': exploitability})

  writer.flush()


if __name__ == '__main__':
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A Least Recently Used cache."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections


class CacheInfo(collections.namedtuple("CacheInfo", [
    "hits", "misses", "size", "max_size"])):
  """Info for LRUCache."""

  @property
  def usage(self):
    return self.size / self.max_size if self.max_size else 0

  @property
  def total(self):
    return self.hits + self.misses

  @property
  def hit_rate(self):
    return self.hits / self.total if self.total else 0


class LRUCache(object):
  """A Least Recently Used cache.

  This is more general than functools.lru_cache since that one requires the
  key to also be the input to the function to generate the value, which
  isn't possible when the input is not hashable, eg a numpy.ndarray.
  """

  def __init__(self, max_size):
    self._max_size = max_size
    self._data = collections.OrderedDict()
    self._hits = 0
    self._misses = 0

  def clear(self):
    self._data.clear()
    self._hits = 0
    self._misses = 0

  def make(self, key, fn):
    """Return the value, either from cache, or make it and save it."""
    try:
      val = self._data.pop(key)  # Take it out.
      self._hits += 1
    except KeyError:
      self._misses += 1
      val = fn()
      if len(self._data) >= self._max_size:
        self._data.popitem(False)
    self._data[key] = val  # Insert/reinsert it at the back.
    return val

  def get(self, key):
    """Get the value and move it to the back, or return None on a miss."""
    try:
      val = self._data.pop(key)  # Take it out.
      self._data[key] = val  # Reinsert it at the back.
      self._hits += 1
      return val
    except KeyError:
      self._misses += 1
      return None

  def set(self, key, val):
    """Set the value."""
    self._data.pop(key, None)  # Take it out if it existed.
    self._data[key] = val  # Insert/reinsert it at the back.
    if len(self._data) > self._max_size:
      self._data.popitem(False)
    return val

  def info(self):
    return CacheInfo(self._hits, self._misses, len(self._data), self._max_size)

  def __len__(self):
    return len(self._data)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python spiel example."""


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pickle

from absl import app
from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from open_spiel.python import games  # pylint: disable=unused-import
from open_spiel.python.algorithms import get_all_states
from open_spiel.python.mfg import games as mfg_games  # pylint:disable=unused-import
import pyspiel
from open_spiel.python.utils import file_utils
# TODO(perolat): add predator_prey in the list of game tested

# Put a bound on length of game so test does not timeout.
MAX_ACTIONS_PER_GAME = 1000

# All games registered in the main spiel library.
SPIEL_GAMES_LIST = pyspiel.registered_games()

# All games loadable without parameter values.
SPIEL_LOADABLE_GAMES_LIST = [g for g in SPIEL_GAMES_LIST if g.default_loadable]

# TODO(b/141950198): Stop hard-coding the number of loadable games.
assert len(SPIEL_LOADABLE_GAMES_LIST) >= 38, len(SPIEL_LOADABLE_GAMES_LIST)

# All simultaneous games.
SPIEL_SIMULTANEOUS_GAMES_LIST = [
    g for g in SPIEL_LOADABLE_GAMES_LIST
    if g.dynamics == pyspiel.GameType.Dynamics.SIMULTANEOUS
]
assert len(SPIEL_SIMULTANEOUS_GAMES_LIST) >= 14, len(
    SPIEL_SIMULTANEOUS_GAMES_LIST)

# All multiplayer games. This is a list of (game, num_players) pairs to test.
SPIEL_MULTIPLAYER_GAMES_LIST = [
    # pylint: disable=g-complex-comprehension
    (g, p)
    for g in SPIEL_LOADABLE_GAMES_LIST
    for p in range(max(g.min_num_players, 2), 1 + min(g.max_num_players, 6))
    if g.max_num_players > 2 and g.max_num_players > g.min_num_players and
    g.short_name != "tiny_hanabi"  # default payoff only works for 2p
    # cannot change the number of players without changing other parameters
    and g.short_name != "universal_poker" and g.short_name != "scotland_yard"
]
assert len(SPIEL_MULTIPLAYER_GAMES_LIST) >= 35, len(
    SPIEL_MULTIPLAYER_GAMES_LIST)


class GamesSimTest(parameterized.TestCase):

  def apply_action(self, state, action):
    if state.is_simultaneous_node():
      assert isinstance(action, list)
      state.apply_actions(action)
    else:
      state.apply_action(action)

  def apply_action_test_clone(self, state, action):
    """Applies the action and tests the clone method if it's implemented."""
    try:
      state_clone = state.clone()
    except Exception:  # pylint: disable=broad-except
      self.apply_action(state, action)
      return
    self.assertEqual(str(state), str(state_clone))
    self.assertEqual(state.history(), state_clone.history())
    self.apply_action(state, action)
    self.apply_action(state_clone, action)
    self.assertEqual(str(state), str(state_clone))
    self.assertEqual(state.history(), state_clone.history())

  def serialize_deserialize(self, game, state, check_pyspiel_serialization,
                            check_pickle_serialization):
    # OpenSpiel native serialization
    if check_pyspiel_serialization:
      ser_str = pyspiel.serialize_game_and_state(game, state)
      new_game, new_state = pyspiel.deserialize_game_and_state(ser_str)
      self.assertEqual(str(game), str(new_game))
      self.assertEqual(str(state), str(new_state))
    if check_pickle_serialization:
      # Pickle serialization + deserialization (of the state).
      pickled_state = pickle.dumps(state)
      unpickled_state = pickle.loads(pickled_state)
      self.assertEqual(str(state), str(unpickled_state))

  def sim_game(
      self,
      game,
      check_pyspiel_serialization=True,
      check_pickle_serialization=True,
      make_distribution_fn=(
          lambda states: ([1 / len(states)] * len(states) if states else []))
  ):
    min_utility = game.min_utility()
    max_utility = game.max_utility()
    self.assertLess(min_utility, max_utility)

    if check_pickle_serialization:
      # Pickle serialization + deserialization (of the game).
      pickled_game = pickle.dumps(game)
      unpickled_game = pickle.loads(pickled_game)
      self.assertEqual(str(game), str(unpickled_game))

      # Pickle serialization + deserialization (of the game type).
      pickled_game_type = pickle.dumps(game.get_type())
      unpickled_game_type = pickle.loads(pickled_game_type)
      self.assertEqual(game.get_type(), unpickled_game_type)

    # Get a new state
    for state in game.new_initial_states():
      total_actions = 0

      next_serialize_check = 1

      while not state.is_terminal() and total_actions <= MAX_ACTIONS_PER_GAME:
        total_actions += 1

        # Serialize/Deserialize is costly. Only do it every power of 2 actions.
        if total_actions >= next_serialize_check:
          self.serialize_deserialize(game, state, check_pyspiel_serialization,
                                     check_pickle_serialization)
          next_serialize_check *= 2

        # The state can be four different types: chance node,
        # mean-field-game node, simultaneous node, or decision node
        if state.is_chance_node():
          # Chance node: sample an outcome
          outcomes = state.chance_outcomes()
          self.assertNotEmpty(outcomes)
          action_list, prob_list = zip(*outcomes)
          action = np.random.choice(action_list, p=prob_list)
          state.apply_action(action)
        elif state.is_simultaneous_node():
          # Simultaneous node: sample actions for all players
          chosen_actions = []
          for pid in range(game.num_players()):
            legal_actions = state.legal_actions(pid)
            action = 0 if not legal_actions else np.random.choice(legal_actions)
            chosen_actions.append(action)
          # Apply the joint action and test cloning states.
          self.apply_action_test_clone(state, chosen_actions)
        elif state.is_mean_field_node():
          self.assertEqual(game.get_type().dynamics,
                           pyspiel.GameType.Dynamics.MEAN_FIELD)
          state.update_distribution(
              make_distribution_fn(state.distribution_support()))
        else:
          self.assertTrue(state.is_player_node())
          # Decision node: sample action for the single current player
          action = np.random.choice(state.legal_actions(state.current_player()))
          # Apply action and test state cloning.
          self.apply_action_test_clone(state, action)

      # Max sure at least one action was made.
      self.assertGreater(total_actions, 0,
                         "No actions taken in sim of " + str(game))

      # Either the game is now done, or the maximum actions has been taken.
      if state.is_terminal():
        # Check there are no legal actions.
        self.assertEmpty(state.legal_actions())
        for player in range(game.num_players()):
          self.assertEmpty(state.legal_actions(player))
        # Print utilities for each player.
        utilities = state.returns()
        # Check that player returns are correct
        for player in range(game.num_players()):
          self.assertEqual(state.player_return(player), utilities[player])
        # Check that each one is in range
        for utility in utilities:
          self.assertGreaterEqual(utility, game.min_utility())
          self.assertLessEqual(utility, game.max_utility())
        print("Sim of game {} terminated with {} total actions. Utilities: {}"
              .format(game, total_actions, utilities))
      else:
        print("Sim of game {} terminated after maximum number of actions {}"
              .format(game, MAX_ACTIONS_PER_GAME))

  @parameterized.named_parameters((game_info.short_name, game_info)
                                  for game_info in SPIEL_LOADABLE_GAMES_LIST)
  def test_game_sim(self, game_info):
    game = pyspiel.load_game(game_info.short_name)
    self.assertLessEqual(game_info.min_num_players, game.num_players())
    self.assertLessEqual(game.num_players(), game_info.max_num_players)
    self.sim_game(game)

  @parameterized.named_parameters(
      (game_info.short_name, game_info)
      for game_info in SPIEL_SIMULTANEOUS_GAMES_LIST)
  def test_simultaneous_game_as_turn_based(self, game_info):
    converted_game = pyspiel.load_game_as_turn_based(game_info.short_name)
    self.sim_game(converted_game)

  @parameterized.named_parameters((f"{p}p_{g.short_name}", g, p)
                                  for g, p in SPIEL_MULTIPLAYER_GAMES_LIST)
  def test_multiplayer_game(self, game_info, num_players):
    if game_info.short_name == "python_mfg_predator_prey":
      reward_matrix = np.ones((num_players, num_players))
      dict_args = {
          "players": num_players,
          "reward_matrix": " ".join(str(v) for v in reward_matrix.flatten())
      }
    else:
      dict_args = {"players": num_players}
    game = pyspiel.load_game(game_info.short_name, dict_args)
    self.sim_game(game)

  def test_breakthrough(self):
    # make a smaller (6x6) board
    game = pyspiel.load_game("breakthrough(rows=6,columns=6)")
    self.sim_game(game)

  def test_pig(self):
    # make a smaller lower win score
    game = pyspiel.load_game("pig(players=2,winscore=15)")
    self.sim_game(game)

  def test_efg_game(self):
    game = pyspiel.load_efg_game(pyspiel.get_sample_efg_data())
    # EFG games loaded directly by string cannot serialize because the game's
    # data cannot be passed in via string parameter.
    for _ in range(0, 100):
      self.sim_game(
          game,
          check_pyspiel_serialization=False,
          check_pickle_serialization=False)
    game = pyspiel.load_efg_game(pyspiel.get_kuhn_poker_efg_data())
    for _ in range(0, 100):
      self.sim_game(
          game,
          check_pyspiel_serialization=False,
          check_pickle_serialization=False)
    # EFG games loaded by file should serialize properly:
    filename = file_utils.find_file(
        "open_spiel/games/efg/sample.efg", 2)
    if filename is not None:
      game = pyspiel.load_game("efg_game(filename=" + filename + ")")
      for _ in range(0, 100):
        self.sim_game(game)
    filename = file_utils.find_file(
        "open_spiel/games/efg/sample.efg", 2)
    if filename is not None:
      game = pyspiel.load_game("efg_game(filename=" + filename + ")")
      for _ in range(0, 100):
        self.sim_game(game)

  def test_backgammon_checker_moves(self):
    game = pyspiel.load_game("backgammon")
    state = game.new_initial_state()
    state.apply_action(0)  # Roll 12 and X starts
    action = state.legal_actions()[0]  # First legal action
    # X has player id 0.
    checker_moves = state.spiel_move_to_checker_moves(0, action)
    print("Checker moves:")
    for i in range(2):
      print("pos {}, num {}, hit? {}".format(checker_moves[i].pos,
                                             checker_moves[i].num,
                                             checker_moves[i].hit))
    action2 = state.checker_moves_to_spiel_move(checker_moves)
    self.assertEqual(action, action2)
    action3 = state.translate_action(0, 0, True)  # 0->2, 0->1
    self.assertEqual(action3, 0)

  def test_backgammon_checker_moves_with_hit_info(self):
    game = pyspiel.load_game("backgammon")
    state = game.new_initial_state()
    while not state.is_terminal():
      if state.is_chance_node():
        outcomes_with_probs = state.chance_outcomes()
        action_list, prob_list = zip(*outcomes_with_probs)
        action = np.random.choice(action_list, p=prob_list)
        state.apply_action(action)
      else:
        legal_actions = state.legal_actions()
        player = state.current_player()
        for action in legal_actions:
          action_str = state.action_to_string(player, action)
          checker_moves = (
              state.augment_with_hit_info(
                  player, state.spiel_move_to_checker_moves(player, action)))
          if checker_moves[0].hit or checker_moves[1].hit:
            self.assertGreaterEqual(action_str.find("*"), 0)
          else:
            self.assertLess(action_str.find("*"), 0)
          if action_str.find("*") > 0:
            self.assertTrue(checker_moves[0].hit or checker_moves[1].hit)
          else:
            self.assertTrue(not checker_moves[0].hit and
                            not checker_moves[1].hit)
        action = np.random.choice(legal_actions)
        state.apply_action(action)

  def test_leduc_get_and_set_private_cards(self):
    game = pyspiel.load_game("leduc_poker")
    state = game.new_initial_state()
    state.apply_action(0)   # give player 0 jack of first suit
    state.apply_action(1)   # give player 1 jack of second suit
    # check that we can retrieve those cards
    print(state)
    private_cards = state.get_private_cards()
    self.assertEqual(private_cards, [0, 1])
    # now give them queens instead, get them again, and check that it worked
    state.set_private_cards([2, 3])
    print(state)
    private_cards = state.get_private_cards()
    self.assertEqual(private_cards, [2, 3])

  @parameterized.parameters(
      {"game_name": "blotto"},
      {"game_name": "goofspiel"},
      {"game_name": "kuhn_poker"},
      {"game_name": "tiny_hanabi"},
      {"game_name": "phantom_ttt"},
      {"game_name": "matrix_rps"},
      {"game_name": "kuhn_poker"},
  )
  def test_restricted_nash_response_test(self, game_name):
    rnr_game = pyspiel.load_game(
        f"restricted_nash_response(game={game_name}())")
    for _ in range(10):
      self.sim_game(rnr_game, check_pyspiel_serialization=False,
                    check_pickle_serialization=False)

# TODO(perolat): find the list of games where it is reasonable to call
# get_all_states
  @parameterized.parameters(
      {"game_name": "python_mfg_crowd_modelling"},
      {"game_name": "mfg_crowd_modelling"},
      {"game_name": "mfg_crowd_modelling_2d"},
      {"game_name": "kuhn_poker"},
      {"game_name": "leduc_poker"},
  )
  def test_has_at_least_an_action(self, game_name):
    """Check that all population's state have at least one action."""
    game = pyspiel.load_game(game_name)
    to_string = lambda s: s.observation_string(pyspiel.PlayerId.
                                               DEFAULT_PLAYER_ID)
    states = get_all_states.get_all_states(
        game,
        depth_limit=-1,
        include_terminals=False,
        include_chance_states=False,
        include_mean_field_states=False,
        to_string=to_string)
    for state in states.values():
      self.assertNotEmpty(state.legal_actions())


def main(_):
  absltest.main()


if __name__ == "__main__":
  # Necessary to run main via app.run for internal tests.
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Representation of a policy for a game.

This is a standard representation for passing policies into algorithms,
with currently the following implementations:

  TabularPolicy - an explicit policy per state, stored in an array
    of shape `(num_states, num_actions)`, convenient for tabular policy
    solution methods.
  UniformRandomPolicy - a uniform distribution over all legal actions for
    the specified player. This is computed as needed, so can be used for
    games where a tabular policy would be unfeasibly large.

The main way of using a policy is to call `action_probabilities(state,
player_id`), to obtain a dict of {action: probability}. `TabularPolicy`
objects expose a lower-level interface, which may be more efficient for
some use cases.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools
from typing import Iterable

import numpy as np

from open_spiel.python.algorithms import get_all_states
import pyspiel


def child(state, action):
  """Returns a child state, handling the simultaneous node case."""
  if isinstance(action, Iterable):
    child_state = state.clone()
    child_state.apply_actions(action)
    return child_state
  else:
    return state.child(action)


def joint_action_probabilities_aux(state, policy):
  """Auxiliary function for joint_action_probabilities.

  Args:
    state: a game state at a simultaneous decision node.
    policy: policy that gives the probability distribution over the legal
      actions for each players.

  Returns:
    actions_per_player: list of list of actions for each player
    probs_per_player: list of list of probabilities do the corresponding action
     in actions_per_player for each player.
  """
  assert state.is_simultaneous_node()
  action_probs_per_player = [
      policy.action_probabilities(state, player)
      for player in range(state.get_game().num_players())
  ]
  actions_per_player = [pi.keys() for pi in action_probs_per_player]
  probs_per_player = [pi.values() for pi in action_probs_per_player]
  return actions_per_player, probs_per_player


def joint_action_probabilities(state, policy):
  """Yields action, probability pairs for a joint policy in simultaneous state.

  Args:
    state: a game state at a simultaneous decision node.
    policy: policy that gives the probability distribution over the legal
      actions for each players.

  Yields:
    (action, probability) pairs. An action is a tuple of individual
      actions for each player of the game. The probability is a single joint
      probability (product of all the individual probabilities).
  """
  actions_per_player, probs_per_player = joint_action_probabilities_aux(
      state, policy)
  for actions, probs in zip(
      itertools.product(*actions_per_player),
      itertools.product(*probs_per_player)):
    yield actions, np.prod(probs)


class Policy:
  """Base class for policies.

  A policy is something that returns a distribution over possible actions
  given a state of the world.

  Attributes:
    game: the game for which this policy applies
    player_ids: list of player ids for which this policy applies; each in the
      interval [0..game.num_players()-1].
  """

  def __init__(self, game, player_ids):
    """Initializes a policy.

    Args:
      game: the game for which this policy applies
      player_ids: list of player ids for which this policy applies; each should
        be in the range 0..game.num_players()-1.
    """
    self.game = game
    self.player_ids = player_ids

  def action_probabilities(self, state, player_id=None):
    """Returns a dictionary {action: prob} for all legal actions.

    IMPORTANT: We assume the following properties hold:
    - All probabilities are >=0 and sum to 1
    - TLDR: Policy implementations should list the (action, prob) for all legal
      actions, but algorithms should not rely on this (yet).
      Details: Before May 2020, only legal actions were present in the mapping,
      but it did not have to be exhaustive: missing actions were considered to
      be associated to a zero probability.
      For example, a deterministic state-poliy was previously {action: 1.0}.
      Given this change of convention is new and hard to enforce, algorithms
      should not rely on the fact that all legal actions should be present.

    Args:
      state: A `pyspiel.State` object.
      player_id: Optional, the player id for whom we want an action. Optional
        unless this is a simultaneous state at which multiple players can act.

    Returns:
      A `dict` of `{action: probability}` for the specified player in the
      supplied state.
    """
    raise NotImplementedError()

  def __call__(self, state, player_id=None):
    """Turns the policy into a callable.

    Args:
      state: The current state of the game.
      player_id: Optional, the player id for whom we want an action. Optional
        unless this is a simultaneous state at which multiple players can act.

    Returns:
      Dictionary of action: probability.
    """
    return self.action_probabilities(state, player_id)

  def to_tabular(self, states=None):
    """Returns a new `TabularPolicy` equivalent to this policy.

    Args:
      states: States of the game that will be used for the tabular policy. If
        None, then get_tabular_policy_states() method will be used to generate
        them.

    Returns:
      a TabularPolicy.
    """
    states = states or get_tabular_policy_states(self.game)
    tabular_policy = TabularPolicy(self.game, self.player_ids, states=states)
    for index, state in enumerate(tabular_policy.states):
      tabular_policy.action_probability_array[index, :] = 0
      for action, probability in self.action_probabilities(state).items():
        tabular_policy.action_probability_array[index, action] = probability
    return tabular_policy


class TabularPolicy(Policy):
  """Policy implementation where the policy is in explicit tabular form.

  In addition to implementing the `Policy` interface, this class exposes
  details of the policy representation for easy manipulation.

  The states are guaranteed to be grouped by player, which can simplify
  code for users of this class, i.e. `action_probability_array` contains
  states for player 0 first, followed by states for player 1, etc.

  The policy uses `state.information_state_string` as the keys if available,
  otherwise `state.observation_string`.

  Usages:

  - Set `policy(info_state, action)`:
  ```
  tabular_policy = TabularPolicy(game)
  info_state_str = state.information_state_string(<optional player>)
  state_policy = tabular_policy.policy_for_key(info_state_str)
  state_policy[action] = <value>
  ```
  - Set `policy(info_state)`:
  ```
  tabular_policy = TabularPolicy(game)
  info_state_str = state.information_state_string(<optional player>)
  state_policy = tabular_policy.policy_for_key(info_state_str)
  state_policy[:] = <list or numpy.array>
  ```

  Attributes:
    action_probability_array: array of shape `(num_states, num_actions)`, where
      `action_probability_array[s, a]` is the probability of choosing action `a`
      when at state `s`.
    state_lookup: `dict` mapping state key string to index into the
      `tabular_policy` array. If information state strings overlap, e.g. for
      different players or if the information state string has imperfect recall,
      then those states will be mapped to the same policy.
    legal_actions_mask: array of shape `(num_states, num_actions)`, each row
      representing which of the possible actions in the game are valid in this
      particular state, containing 1 for valid actions, 0 for invalid actions.
    states_per_player: A `list` per player of the state key strings at which
      they have a decision to make.
    states: A `list` of the states as ordered in the `action_probability_array`.
    state_in: array of shape `(num_states, state_vector_size)` containing the
      normalised vector representation of each information state. Populated only
      for games which support information_state_tensor(), and is None otherwise.
    game_type: The game attributes as returned by `Game::GetType`; used to
      determine whether to use information state or observation as the key in
      the tabular policy.
  """

  def __init__(self,
               game,
               players=None,
               to_string=lambda s: s.history_str(),
               states=None):
    """Initializes a uniform random policy for all players in the game."""
    players = sorted(players or range(game.num_players()))
    super().__init__(game, players)
    self.game_type = game.get_type()

    # Get all states in the game at which players have to make decisions unless
    # they are explicitly specified.
    states = states or get_all_states.get_all_states(
        game,
        depth_limit=-1,
        include_terminals=False,
        include_chance_states=False,
        include_mean_field_states=False,
        to_string=to_string)

    # Assemble legal actions for every valid (state, player) pair, keyed by
    # information state string.
    self.state_lookup = {}
    self.states_per_player = [[] for _ in range(game.num_players())]
    self.states = []
    legal_actions_list = []
    state_in_list = []
    for player in players:
      # States are ordered by their history.
      for _, state in sorted(states.items(), key=lambda pair: pair[0]):
        if state.is_simultaneous_node() or player == state.current_player():
          legal_actions = state.legal_actions_mask(player)
          if any(legal_actions):
            key = self._state_key(state, player)
            if key not in self.state_lookup:
              state_index = len(legal_actions_list)
              self.state_lookup[key] = state_index
              legal_actions_list.append(legal_actions)
              self.states_per_player[player].append(key)
              self.states.append(state)
              if self.game_type.provides_information_state_tensor:
                state_in_list.append(state.information_state_tensor(player))
              elif self.game_type.provides_observation_tensor:
                state_in_list.append(state.observation_tensor(player))

    # Put legal action masks in a numpy array and create the uniform random
    # policy.
    self.state_in = None
    if state_in_list:
      self.state_in = np.array(state_in_list)
    self.legal_actions_mask = np.array(legal_actions_list)
    self.action_probability_array = (
        self.legal_actions_mask /
        np.sum(self.legal_actions_mask, axis=-1, keepdims=True))

  def _state_key(self, state, player):
    """Returns the key to use to look up this (state, player) pair."""
    if self.game_type.provides_information_state_string:
      if player is None:
        return state.information_state_string()
      return state.information_state_string(player)
    if self.game_type.provides_observation_string:
      if player is None:
        return state.observation_string()
      return state.observation_string(player)
    return str(state)

  def action_probabilities(self, state, player_id=None):
    """Returns an {action: probability} dict, covering all legal actions."""
    legal_actions = (
        state.legal_actions()
        if player_id is None else state.legal_actions(player_id))
    if not legal_actions:
      return {0: 1.0}
    probability = self.policy_for_key(self._state_key(state, player_id))
    return {action: probability[action] for action in legal_actions}

  def state_index(self, state):
    """Returns the index in the TabularPolicy associated to `state`."""
    return self.state_lookup[self._state_key(state, state.current_player())]

  def policy_for_key(self, key):
    """Returns the policy as a vector given a state key string.

    Args:
      key: A key for the specified state.

    Returns:
      A vector of probabilities, one per action. This is a slice of the
      backing policy array, and so slice or index assignment will update the
      policy. For example:
      ```
      tabular_policy.policy_for_key(s)[:] = [0.1, 0.5, 0.4]
      ```
    """
    return self.action_probability_array[self.state_lookup[key]]

  def __copy__(self, copy_action_probability_array=True):
    """Returns a shallow copy of self.

    Most class attributes will be pointers to the copied object's attributes,
    and therefore altering them could lead to unexpected behavioural changes.
    Only action_probability_array is expected to be modified.

    Args:
      copy_action_probability_array: Whether to also include
        action_probability_array in the copy operation.

    Returns:
      Copy.
    """
    result = TabularPolicy.__new__(TabularPolicy)
    result.state_lookup = self.state_lookup
    result.game_type = self.game_type
    result.legal_actions_mask = self.legal_actions_mask
    result.state_in = self.state_in
    result.state_lookup = self.state_lookup
    result.states_per_player = self.states_per_player
    result.states = self.states
    result.game = self.game
    result.player_ids = self.player_ids
    if copy_action_probability_array:
      result.action_probability_array = np.copy(self.action_probability_array)
    return result

  def copy_with_noise(self,
                      alpha=0.0,
                      beta=0.0,
                      random_state=np.random.RandomState()):
    """Returns a copy of this policy perturbed with noise.

    Generates a new random distribution using a softmax on normal random
    variables with temperature beta, and mixes it with the old distribution
    using 1-alpha * old_distribution + alpha * random_distribution.
    Args:
      alpha: Parameter characterizing the mixture amount between new and old
        distributions. Between 0 and 1.
        alpha = 0: keep old table.
        alpha = 1: keep random table.
      beta: Temperature of the softmax. Makes for more extreme policies.
      random_state: A numpy `RandomState` object. If not provided, a shared
        random state will be used.

    Returns:
      Perturbed copy.
    """
    copied_instance = self.__copy__(False)
    probability_array = self.action_probability_array
    noise_mask = random_state.normal(size=probability_array.shape)
    noise_mask = np.exp(beta * noise_mask) * self.legal_actions_mask
    noise_mask = noise_mask / (np.sum(noise_mask, axis=1).reshape(-1, 1))
    copied_instance.action_probability_array = (
        1 - alpha) * probability_array + alpha * noise_mask
    return copied_instance


class UniformRandomPolicy(Policy):
  """Policy where the action distribution is uniform over all legal actions.

  This is computed as needed, so can be used for games where a tabular policy
  would be unfeasibly large, but incurs a legal action computation every time.
  """

  def __init__(self, game):
    """Initializes a uniform random policy for all players in the game."""
    all_players = list(range(game.num_players()))
    super().__init__(game, all_players)

  def action_probabilities(self, state, player_id=None):
    """Returns a uniform random policy for a player in a state.

    Args:
      state: A `pyspiel.State` object.
      player_id: Optional, the player id for which we want an action. Optional
        unless this is a simultaneous state at which multiple players can act.

    Returns:
      A `dict` of `{action: probability}` for the specified player in the
      supplied state. This will contain all legal actions, each with the same
      probability, equal to 1 / num_legal_actions.
    """
    legal_actions = (
        state.legal_actions()
        if player_id is None else state.legal_actions(player_id))
    if not legal_actions:
      return {0: 1.0}
    probability = 1 / len(legal_actions)
    return {action: probability for action in legal_actions}


class FirstActionPolicy(Policy):
  """A policy that always takes the lowest-numbered legal action."""

  def __init__(self, game):
    all_players = list(range(game.num_players()))
    super().__init__(game, all_players)

  def action_probabilities(self, state, player_id=None):
    legal_actions = (
        state.legal_actions()
        if player_id is None else state.legal_actions(player_id))
    if not legal_actions:
      return {0: 1.0}
    min_action = min(legal_actions)
    return {
        action: 1.0 if action == min_action else 0.0 for action in legal_actions
    }


def get_tabular_policy_states(game):
  """Returns the states of the game for a tabular policy."""
  if game.get_type().dynamics == pyspiel.GameType.Dynamics.MEAN_FIELD:
    # TODO(perolat): We use s.observation_string(DEFAULT_MFG_PLAYER) here as the
    # number of history is exponential on the depth of the MFG. What we really
    # need is a representation of the state. For many player Mean Field games,
    # the state will be (x0, x1, x2, ..., xn) and the observation_string(0) will
    # output the string of x0. In that case we would need something like
    # str([observation_string(i) for i in range(num_player)])
    to_string = lambda s: s.observation_string(pyspiel.PlayerId.
                                               DEFAULT_PLAYER_ID)
  else:
    to_string = lambda s: s.history_str()
  return get_all_states.get_all_states(
      game,
      depth_limit=-1,
      include_terminals=False,
      include_chance_states=False,
      include_mean_field_states=False,
      to_string=to_string)


def tabular_policy_from_callable(game, callable_policy, players=None):
  """Converts a legacy callable policy into a TabularPolicy.

  Recommendation - instead of using this to convert your policy for evaluation
  purposes, work directly with a `TabularPolicy` if possible.
  Second choice - work with a `Policy` class and call `to_tabular` as needed.

  Args:
    game: The game for which we want a TabularPolicy.
    callable_policy: A callable: state -> action probabilities dict or list.
    players: List of players this policy applies to. If `None`, applies to all
      players.

  Returns:
    A TabularPolicy that materializes the callable policy.
  """
  tabular_policy = TabularPolicy(game, players)
  for state_index, state in enumerate(tabular_policy.states):
    action_probabilities = dict(callable_policy(state))
    infostate_policy = [
        action_probabilities.get(action, 0.)
        for action in range(game.num_distinct_actions())
    ]
    tabular_policy.action_probability_array[state_index, :] = infostate_policy
  return tabular_policy


def pyspiel_policy_to_python_policy(game, pyspiel_tabular_policy, players=None):
  """Converts a pyspiel.TabularPolicy to a TabularPolicy.

  Args:
    game: The OpenSpiel game.
    pyspiel_tabular_policy: Pyspiel tabular policy to copy from.
    players: List of integer player ids to copy policy from. For example,
      `players=[0]` will only copy player 0's policy over into the python policy
      (the other player's policies will be undefined). Default value of `None`
      will copy all players' policies.

  Returns:
    python_policy
  """
  policy = TabularPolicy(game, players=players)
  for item in pyspiel_tabular_policy.policy_table().items():
    info_state_str, actions_probs = item
    # If requested, only populate a policy for particular players.
    if players is not None and info_state_str not in policy.state_lookup:
      continue
    state_policy = policy.policy_for_key(info_state_str)
    for action, prob in actions_probs:
      state_policy[action] = prob
  return policy


def python_policy_to_pyspiel_policy(python_tabular_policy):
  """Converts a TabularPolicy to a pyspiel.TabularPolicy."""
  infostates_to_probabilities = dict()
  for infostate, index in python_tabular_policy.state_lookup.items():
    probs = python_tabular_policy.action_probability_array[index]
    legals = python_tabular_policy.legal_actions_mask[index]

    action_probs = []
    for action, (prob, is_legal) in enumerate(zip(probs, legals)):
      if is_legal == 1:
        action_probs.append((action, prob))
    infostates_to_probabilities[infostate] = action_probs
  return pyspiel.TabularPolicy(infostates_to_probabilities)


def python_policies_to_pyspiel_policies(policies):
  """Same conversion as above (list version).

  Args:
    policies: a list of python.TabularPolicy

  Returns:
    a list of pyspiel.TabularPolicy.
  """
  return [python_policy_to_pyspiel_policy(p) for p in policies]


def merge_tabular_policies(tabular_policies, game):
  """Merges n_player policies into single joint policy.

  Missing states are filled with a valid uniform policy.

  Args:
    tabular_policies: List of python TabularPolicy (one for each player).
    game: The game corresponding to the resulting TabularPolicy.

  Returns:
    merged_policy: A TabularPolicy with each player i's policy taken from the
      ith joint_policy.
  """
  if len(tabular_policies) != game.num_players():
    raise ValueError("len(tabular_policies) != num_players: %d != %d" %
                     (len(tabular_policies), game.num_players()))
  merged_policy = TabularPolicy(game)
  for p, p_states in enumerate(merged_policy.states_per_player):
    for p_state in p_states:
      to_index = merged_policy.state_lookup[p_state]
      # Only copy if the state exists, otherwise fall back onto uniform.
      if p_state in tabular_policies[p].state_lookup:
        from_index = tabular_policies[p].state_lookup[p_state]
        merged_policy.action_probability_array[to_index] = (
            tabular_policies[p].action_probability_array[from_index])
  return merged_policy



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.action_value_vs_best_response.py."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest

import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import action_value_vs_best_response
import pyspiel


class ActionValuesVsBestResponseTest(absltest.TestCase):

  def test_kuhn_poker_uniform(self):
    game = pyspiel.load_game("kuhn_poker")
    calc = action_value_vs_best_response.Calculator(game)
    (expl, avvbr, cfrp,
     player_reach_probs) = calc(0, policy.UniformRandomPolicy(game),
                                ["0", "1", "2", "0pb", "1pb", "2pb"])
    self.assertAlmostEqual(expl, 15 / 36)
    np.testing.assert_allclose(
        avvbr,
        [
            [-1.5, -2.0],  # 0 (better to pass)
            [-0.5, -0.5],  # 1 (same)
            [0.5, 1.5],  # 2 (better to bet)
            [-1.0, -2.0],  # 0pb - losing
            [-1.0, 0.0],  # 1pb - best response is bet always
            [-1.0, 2.0],  # 2pb - winning
        ])
    np.testing.assert_allclose(cfrp, [1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3])
    np.testing.assert_allclose([1, 1, 1, 1 / 2, 1 / 2, 1 / 2],
                               player_reach_probs)

  def test_kuhn_poker_always_pass_p0(self):
    game = pyspiel.load_game("kuhn_poker")
    calc = action_value_vs_best_response.Calculator(game)
    (expl, avvbr, cfrp, player_reach_probs) = calc(
        0, policy.FirstActionPolicy(game),
        ["0", "1", "2", "0pb", "1pb", "2pb"])
    self.assertAlmostEqual(expl, 1.)
    np.testing.assert_allclose(
        avvbr,
        [
            # Opening bet. If we pass, we always lose (pass-pass with op's K,
            # otherwise pass-bet-pass).
            # If we bet, we always win (because op's best response is to pass,
            # because this is an unreachable state and we break ties in favour
            # of the lowest action).
            [-1, 1],
            [-1, 1],
            [-1, 1],
            # We pass, opp bets into us. This can be either J or Q (K will pass
            # because of the tie-break rules).
            # So we are guaranteed to be winning with Q or K.
            [-1, -2],  # 0pb
            [-1, 2],  # 1pb
            [-1, 2],  # 2pb
        ])
    np.testing.assert_allclose(cfrp, [1 / 3, 1 / 3, 1 / 3, 1 / 6, 1 / 6, 1 / 3])
    np.testing.assert_allclose([1., 1., 1., 1., 1., 1.], player_reach_probs)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mirror Descent (https://arxiv.org/pdf/2103.00623.pdf)."""
from typing import Optional

import numpy as np

from open_spiel.python import policy as policy_std
from open_spiel.python.mfg import value
from open_spiel.python.mfg.algorithms import distribution
import pyspiel


def softmax_projection(logits):
  max_l = max(logits)
  exp_l = [np.exp(l - max_l) for l in logits]
  norm_exp = sum(exp_l)
  return [l / norm_exp for l in exp_l]


class ProjectedPolicy(policy_std.Policy):
  """Project values on the policy simplex."""

  def __init__(self, game, player_ids,
               cumulative_state_value: value.ValueFunction):
    """Initializes the projected policy.

    Args:
      game: The game to analyze.
      player_ids: list of player ids for which this policy applies; each should
        be in the range 0..game.num_players()-1.
      cumulative_state_value: The cumulative state value to project.
    """
    super(ProjectedPolicy, self).__init__(game, player_ids)
    self._cumulative_state_value = cumulative_state_value

  def cumulative_value(self, state, action=None):
    if action is None:
      return self._cumulative_state_value(
          state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID))
    else:
      new_state = state.child(action)
      return state.rewards()[0] + self._cumulative_state_value(
          new_state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID))

  def action_probabilities(self, state, player_id=None):
    action_logit = [(a, self.cumulative_value(state, action=a))
                    for a in state.legal_actions()]
    action, logit = zip(*action_logit)
    prob = softmax_projection(logit)
    action_prob = zip(action, prob)
    return dict(action_prob)


class MirrorDescent(object):
  """The mirror descent algorithm."""

  def __init__(self,
               game,
               state_value: Optional[value.ValueFunction] = None,
               lr=0.01,
               root_state=None):
    """Initializes mirror descent.

    Args:
      game: The game,
      state_value: A state value function. Default to TabularValueFunction.
      lr: The learning rate of mirror descent,
      root_state: The state of the game at which to start. If `None`, the game
        root state is used.
    """
    self._game = game
    if root_state is None:
      self._root_states = game.new_initial_states()
    else:
      self._root_states = [root_state]
    self._policy = policy_std.UniformRandomPolicy(game)
    self._distribution = distribution.DistributionPolicy(game, self._policy)
    self._md_step = 0
    self._lr = lr

    self._state_value = (state_value if state_value
                         else value.TabularValueFunction(game))
    self._cumulative_state_value = value.TabularValueFunction(game)

  def eval_state(self, state, learning_rate):
    """Evaluate the value of a state and update the cumulative sum."""
    state_str = state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID)
    if self._state_value.has(state_str):
      return self._state_value(state_str)
    elif state.is_terminal():
      self._state_value.set_value(
          state_str,
          state.rewards()[state.mean_field_population()])
      self._cumulative_state_value.add_value(
          state_str, learning_rate * self._state_value(state_str))
      return self._state_value(state_str)
    elif state.current_player() == pyspiel.PlayerId.CHANCE:
      self._state_value.set_value(state_str, 0.0)
      for action, prob in state.chance_outcomes():
        new_state = state.child(action)
        self._state_value.add_value(
            state_str, prob * self.eval_state(new_state, learning_rate))
      self._cumulative_state_value.add_value(
          state_str, learning_rate * self._state_value(state_str))
      return self._state_value(state_str)
    elif state.current_player() == pyspiel.PlayerId.MEAN_FIELD:
      dist_to_register = state.distribution_support()
      dist = [
          self._distribution.value_str(str_state, 0.0)
          for str_state in dist_to_register
      ]
      new_state = state.clone()
      new_state.update_distribution(dist)
      self._state_value.set_value(
          state_str,
          state.rewards()[state.mean_field_population()] +
          self.eval_state(new_state, learning_rate))
      self._cumulative_state_value.add_value(
          state_str, learning_rate * self._state_value(state_str))
      return self._state_value(state_str)
    else:
      assert int(state.current_player()) >= 0, "The player id should be >= 0"
      v = 0.0
      for action, prob in self._policy.action_probabilities(state).items():
        new_state = state.child(action)
        v += prob * self.eval_state(new_state, learning_rate)
      self._state_value.set_value(
          state_str,
          state.rewards()[state.mean_field_population()] + v)
      self._cumulative_state_value.add_value(
          state_str, learning_rate * self._state_value(state_str))
      return self._state_value(state_str)

  def iteration(self, learning_rate=None):
    """an iteration of Mirror Descent."""
    self._md_step += 1
    # TODO(sertan): Fix me.
    self._state_value = value.TabularValueFunction(self._game)
    for state in self._root_states:
      self.eval_state(state, learning_rate if learning_rate else self._lr)
    self._policy = ProjectedPolicy(self._game,
                                   list(range(self._game.num_players())),
                                   self._cumulative_state_value)
    self._distribution = distribution.DistributionPolicy(
        self._game, self._policy)

  def get_policy(self):
    return self._policy



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.referee."""

import os
from absl import flags
from absl.testing import absltest
import pyspiel


flags.DEFINE_string("bot_dir",
                    os.path.dirname(__file__) + "/../bots",
                    "Path to python implementation of bots.")
FLAGS = flags.FLAGS


class RefereeTest(absltest.TestCase):

  def test_playing_tournament(self):
    ref = pyspiel.Referee(
        "kuhn_poker", [f"python {FLAGS.bot_dir}/higc_random_bot_test.py"] * 2,
        settings=pyspiel.TournamentSettings(
            timeout_ready=2000, timeout_start=500))
    results = ref.play_tournament(num_matches=1)
    self.assertLen(results.matches, 1)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Export game in gambit .efg format."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags
from absl import logging

from open_spiel.python.algorithms.gambit import export_gambit
import pyspiel

FLAGS = flags.FLAGS
flags.DEFINE_string("game", "kuhn_poker", "Name of the game")
flags.DEFINE_string("out", "/tmp/gametree.efg", "Name of output file, e.g., "
                    "[*.efg].")
flags.DEFINE_boolean("print", False, "Print the tree to stdout "
                     "instead of saving to file.")


def main(argv):
  del argv

  game = pyspiel.load_game(FLAGS.game)
  game_type = game.get_type()

  if game_type.dynamics == pyspiel.GameType.Dynamics.SIMULTANEOUS:
    logging.warn("%s is not turn-based. Trying to reload game as turn-based.",
                 FLAGS.game)
    game = pyspiel.load_game_as_turn_based(FLAGS.game)

  gametree = export_gambit(game)  # use default decorators
  if FLAGS.print:
    print(gametree)
  else:
    with open(FLAGS.out, "w") as f:
      f.write(gametree)
    logging.info("Game tree for %s saved to file: %s", FLAGS.game, FLAGS.out)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit test for the HIGC random bot.

This test mimics the basic C++ tests in higc/bots/random_bot.py and is
duplicated here to make automated wheels tests work in the absence
of the higc/ directory.
"""

import base64
import sys
import numpy as np
from open_spiel.python.observation import make_observation
import pyspiel


game_name = input()
play_as = int(input())
game = pyspiel.load_game(game_name)

public_observation = make_observation(
    game,
    pyspiel.IIGObservationType(
        perfect_recall=False,
        public_info=True,
        private_info=pyspiel.PrivateInfoType.NONE))
private_observation = make_observation(
    game,
    pyspiel.IIGObservationType(
        perfect_recall=False,
        public_info=False,
        private_info=pyspiel.PrivateInfoType.SINGLE_PLAYER))
print("ready")

while True:
  print("start")
  state = game.new_initial_state()
  while True:
    message = input()
    if message == "tournament over":
      print("tournament over")
      sys.exit(0)

    if message.startswith("match over"):
      print("match over")
      break

    public_buf, private_buf, *legal_actions = message.split(" ")
    public_observation.decompress(base64.b64decode(public_buf))
    private_observation.decompress(base64.b64decode(private_buf))

    if legal_actions:
      print(np.random.choice(legal_actions))
    else:
      print("ponder")

  assert message.startswith("match over")
  score = int(message.split(" ")[-1])
  print("score:", score, file=sys.stderr)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.egt.alpharank."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest

# pylint: disable=g-import-not-at-top
import matplotlib
matplotlib.use("agg")  # switch backend for testing

import numpy as np

from open_spiel.python.egt import alpharank
from open_spiel.python.egt import heuristic_payoff_table
from open_spiel.python.egt import utils
import pyspiel


class AlphaRankTest(absltest.TestCase):

  def test_stationary_distribution(self):
    """Tests stationary distribution using payoffs from Han et al., 2013."""
    r = 1.
    t = 2.
    p = 0.
    s = -1.
    delta = 4.
    eps = 0.25
    payoff_tables = [
        np.asarray([[r - eps / 2., r - eps, 0, s + delta - eps, r - eps],
                    [r, r, s, s, s], [0, t, p, p, p], [t - delta, t, p, p, p],
                    [r, t, p, p, p]])
    ]

    m = 20
    alpha = 0.1
    expected_pi = np.asarray(
        [0.40966787, 0.07959841, 0.20506998, 0.08505983, 0.2206039])

    # Test payoffs in matrix format
    _, _, pi_matrix, _, _ = alpharank.compute(
        payoff_tables, m=m, alpha=alpha, use_local_selection_model=False)
    np.testing.assert_array_almost_equal(pi_matrix, expected_pi, decimal=4)

    # Test payoffs in HPT format
    hpts = [heuristic_payoff_table.from_matrix_game(payoff_tables[0])]
    _, _, pi_hpts, _, _ = alpharank.compute(
        hpts, m=m, alpha=alpha, use_local_selection_model=False)
    np.testing.assert_array_almost_equal(pi_hpts, expected_pi, decimal=4)

  def test_constant_sum_transition_matrix(self):
    """Tests closed-form transition matrix computation for constant-sum case."""

    game = pyspiel.load_matrix_game("matrix_rps")
    payoff_tables = utils.game_payoffs_array(game)

    # Checks if the game is symmetric and runs single-population analysis if so
    _, payoff_tables = utils.is_symmetric_matrix_game(payoff_tables)
    payoffs_are_hpt_format = utils.check_payoffs_are_hpt(payoff_tables)

    m = 20
    alpha = 0.1

    # Case 1) General-sum game computation (slower)
    game_is_constant_sum = False
    use_local_selection_model = False
    payoff_sum = None
    c1, rhos1 = alpharank._get_singlepop_transition_matrix(
        payoff_tables[0], payoffs_are_hpt_format, m, alpha,
        game_is_constant_sum, use_local_selection_model, payoff_sum)

    # Case 2) Constant-sum closed-form computation (faster)
    game_is_constant_sum, payoff_sum = utils.check_is_constant_sum(
        payoff_tables[0], payoffs_are_hpt_format)
    c2, rhos2 = alpharank._get_singlepop_transition_matrix(
        payoff_tables[0], payoffs_are_hpt_format, m, alpha,
        game_is_constant_sum, use_local_selection_model, payoff_sum)

    # Ensure both cases match
    np.testing.assert_array_almost_equal(c1, c2)
    np.testing.assert_array_almost_equal(rhos1, rhos2)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example of ResponseGraphUCB run on a 2x2 game."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
import matplotlib.pyplot as plt
import numpy as np

from open_spiel.python.algorithms import response_graph_ucb
from open_spiel.python.algorithms import response_graph_ucb_utils


def get_example_2x2_payoffs():
  mean_payoffs = np.random.uniform(-1, 1, size=(2, 2, 2))
  mean_payoffs[0, :, :] = np.asarray([[0.5, 0.85], [0.15, 0.5]])
  mean_payoffs[1, :, :] = 1 - mean_payoffs[0, :, :]
  return mean_payoffs


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  mean_payoffs = get_example_2x2_payoffs()
  game = response_graph_ucb_utils.BernoulliGameSampler(
      [2, 2], mean_payoffs, payoff_bounds=[-1., 1.])
  game.p_max = mean_payoffs
  game.means = mean_payoffs
  print('Game means:\n', game.means)

  exploration_strategy = 'uniform-exhaustive'
  confidence_method = 'ucb-standard'
  r_ucb = response_graph_ucb.ResponseGraphUCB(
      game,
      exploration_strategy=exploration_strategy,
      confidence_method=confidence_method,
      delta=0.1)
  results = r_ucb.run()

  # Plotting
  print('Number of total samples: {}'.format(np.sum(r_ucb.count[0])))
  r_ucb.visualise_2x2x2(real_values=game.means, graph=results['graph'])
  r_ucb.visualise_count_history(figsize=(5, 3))
  plt.gca().xaxis.label.set_fontsize(15)
  plt.gca().yaxis.label.set_fontsize(15)

  # Compare to ground truth graph
  real_graph = r_ucb.construct_real_graph()
  r_ucb.plot_graph(real_graph)
  plt.show()

if __name__ == '__main__':
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python XFP example."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
from absl import app
from absl import flags

from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import fictitious_play
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_integer("iterations", 100, "Number of iterations")
flags.DEFINE_string("game", "leduc_poker", "Name of the game")
flags.DEFINE_integer("players", 2, "Number of players")
flags.DEFINE_integer("print_freq", 10, "How often to print the exploitability")


def main(_):
  game = pyspiel.load_game(FLAGS.game, {"players": FLAGS.players})
  xfp_solver = fictitious_play.XFPSolver(game)
  for i in range(FLAGS.iterations):
    xfp_solver.iteration()
    conv = exploitability.exploitability(game, xfp_solver.average_policy())
    if i % FLAGS.print_freq == 0:
      print("Iteration: {} Conv: {}".format(i, conv))
      sys.stdout.flush()


if __name__ == "__main__":
  app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Factory to create (benchmark) MFG games with different settings."""

from typing import Optional

from absl import logging

from open_spiel.python.games import dynamic_routing_data
import open_spiel.python.mfg.games as games  # pylint: disable=unused-import
from open_spiel.python.mfg.games import crowd_modelling_2d
from open_spiel.python.mfg.games import dynamic_routing
import pyspiel

# For each game, the setting with the game name, e.g. python_mfg_dynamic_routing
# for dynamic routing, denotes the default parameters. Variations are not
# prefixed by the exact game name so that they can be used with different
# implementations, e.g. Python or C++, of the same game. Empty parameters use
# the default values as specified in the game.
GAME_SETTINGS = {
    # 2D crowd modelling game.
    "crowd_modelling_2d_10x10": {},
    "crowd_modelling_2d_four_rooms": {
        **crowd_modelling_2d.FOUR_ROOMS,
        "only_distribution_reward": True,
    },
    "crowd_modelling_2d_maze": {
        **crowd_modelling_2d.MAZE,
        "only_distribution_reward": True,
    },
    # Dynamic routing game.
    "dynamic_routing_braess": {
        "max_num_time_step": 100,
        "network": "braess",
        "time_step_length": 0.05,
    },
    "dynamic_routing_line": {
        "max_num_time_step": 5,
        "network": "line",
        "time_step_length": 1.0,
    },
    "dynamic_routing_sioux_falls_dummy_demand": {
        "max_num_time_step": 81,
        "network": "sioux_falls_dummy_demand",
        "time_step_length": 0.5,
    },
    "dynamic_routing_sioux_falls": {
        # TODO(cabannes): change these values based on experiment output.
        "max_num_time_step": 81,
        "network": "sioux_falls",
        "time_step_length": 0.5,
    },
    # Predator and prey game.
    "predator_prey_5x5x3": {},
}

# Default settings for the games.
GAME_SETTINGS.update({
    "mfg_crowd_modelling_2d": GAME_SETTINGS["crowd_modelling_2d_10x10"],
    "python_mfg_dynamic_routing": GAME_SETTINGS["dynamic_routing_line"],
    "python_mfg_predator_prey": GAME_SETTINGS["predator_prey_5x5x3"],
})

_DYNAMIC_ROUTING_NETWORK = {
    "line": (dynamic_routing_data.LINE_NETWORK,
             dynamic_routing_data.LINE_NETWORK_OD_DEMAND),
    "braess": (dynamic_routing_data.BRAESS_NETWORK,
               dynamic_routing_data.BRAESS_NETWORK_OD_DEMAND),
    "sioux_falls_dummy_demand":
        (dynamic_routing_data.SIOUX_FALLS_NETWORK,
         dynamic_routing_data.SIOUX_FALLS_DUMMY_OD_DEMAND),
    "sioux_falls": (dynamic_routing_data.SIOUX_FALLS_NETWORK,
                    dynamic_routing_data.SIOUX_FALLS_OD_DEMAND)
}


def create_game_with_setting(game_name: str,
                             setting: Optional[str] = None) -> pyspiel.Game:
  """Creates an OpenSpiel game with the specified setting.

  Args:
    game_name: Name of a registered game, e.g. mfg_crowd_modelling_2d.
    setting: Name of the pre-defined setting. If None, game_name will be used
      instead. The setting should be present in the GAME_SETTINGS map above.

  Returns:
    a Game.
  """
  setting = setting or game_name
  params = GAME_SETTINGS.get(setting)
  if params is None:
    raise ValueError(f"{setting} setting does not exist for {game_name}.")

  logging.info("Creating %s game with parameters: %r", game_name, params)

  # Dynamic routing game requires setting the network and demand explicitly.
  if game_name == "python_mfg_dynamic_routing":
    # Create a copy since we modify it below removing the network key.
    params = params.copy()
    network = params.pop("network")
    network, od_demand = _DYNAMIC_ROUTING_NETWORK[network]
    return dynamic_routing.MeanFieldRoutingGame(
        params, network=network, od_demand=od_demand)

  return pyspiel.load_game(game_name, params)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Computes the value of a given policy."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import List, Union

import numpy as np

from open_spiel.python import policy


def _transitions(state, policies):
  """Returns iterator over (action, prob) from the given state."""
  if state.is_chance_node():
    return state.chance_outcomes()
  elif state.is_simultaneous_node():
    return policy.joint_action_probabilities(state, policies)
  else:
    player = state.current_player()
    return policies[player].action_probabilities(state).items()


def policy_value(state,
                 policies: Union[List[policy.Policy], policy.Policy],
                 probability_threshold: float = 0):
  """Returns the expected values for the state for players following `policies`.

  Computes the expected value of the`state` for each player, assuming player `i`
  follows the policy given in `policies[i]`.

  Args:
    state: A `pyspiel.State`.
    policies: A `list` of `policy.Policy` objects, one per player for sequential
      games, one policy for simulatenous games.
    probability_threshold: only sum over entries with prob greater than this
      (default: 0).

  Returns:
    A `numpy.array` containing the expected value for each player.
  """
  if state.is_terminal():
    return np.array(state.returns())
  else:
    return sum(prob * policy_value(policy.child(state, action), policies)
               for action, prob in _transitions(state, policies)
               if prob > probability_threshold)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.bots.bluechip_bridge_uncontested_bidding."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from open_spiel.python.bots import bluechip_bridge
import pyspiel


class BluechipBridgeWrapperTest(absltest.TestCase):

  def test_complete_deal_east(self):
    # Plays a complete deal, with the mock external bot playing East.
    # The deal is as follows:
    #
    # Vul: None
    #         S AKJ8
    #         H 4
    #         D JT9532
    #         C 32
    # S 3             S Q9
    # H KQJ8762       H AT5
    # D K4            D A87
    # C KQ4           C AJT96
    #         S T76542
    #         H 93
    #         D Q6
    #         C 875
    #
    # West  North East  South
    #       Pass  1N    Pass
    # 2D    Pass  2H    Pass
    # 3S    Dbl   4C    Pass
    # 4D    Pass  4N    Pass
    # 5D    Pass  6H    Pass
    # Pass  Pass
    #
    # N  E  S  W  N  E  S
    #       S7 S3 SK S9
    # DJ D8 D6 DK
    #          H2 H4 HT H9
    #    H5 H3 H6 C3
    #          C4 C2 CT C5
    #    C6 C7 CQ D2
    #          CK D3 CJ C8
    #          D4 D5 DA DQ
    #    C9 S2 H7 S8
    #          HK SJ HA S4
    #    CA S5 H8 D9
    #          HQ DT D7 S6
    #          HJ SA SQ ST
    #
    # Declarer tricks: 12

    game = pyspiel.load_game('bridge(use_double_dummy_result=false)')
    mock_client = absltest.mock.Mock(
        **{
            'read_line.side_effect': [
                'Connecting "WBridge5" as ANYPL using protocol version 18',
                'EAST ready for teams',
                'EAST ready to start',
                'EAST ready for deal',
                'EAST ready for cards',
                "EAST ready for NORTH's bid",
                'EAST bids 1NT',
                "EAST ready for SOUTH's bid",
                "EAST ready for WEST's bid",
                "EAST ready for NORTH's bid",
                'EAST bids 2H',
                "EAST ready for SOUTH's bid",
                "EAST ready for WEST's bid",
                "EAST ready for NORTH's bid",
                'EAST bids 4C Alert.',
                "EAST ready for SOUTH's bid",
                "EAST ready for WEST's bid",
                "EAST ready for NORTH's bid",
                'EAST bids 4NT',
                "EAST ready for SOUTH's bid",
                "EAST ready for WEST's bid",
                "EAST ready for NORTH's bid",
                'EAST bids 6H',
                "EAST ready for SOUTH's bid",
                "EAST ready for WEST's bid",
                "EAST ready for NORTH's bid",
                "EAST ready for SOUTH's card to trick 1",
                'EAST ready for dummy',
                'WEST plays 3s',
                "EAST ready for NORTH's card to trick 1",
                'EAST plays 9s',
                "EAST ready for NORTH's card to trick 2",
                'EAST plays 8d',
                "EAST ready for SOUTH's card to trick 2",
                'WEST plays kd',
                'WEST plays 2h',
                "EAST ready for NORTH's card to trick 3",
                'EAST plays th',
                "EAST ready for SOUTH's card to trick 3",
                'EAST plays 5h',
                "EAST ready for SOUTH's card to trick 4",
                'WEST plays 6h',
                "EAST ready for NORTH's card to trick 4",
                'WEST plays 4c',
                "EAST ready for NORTH's card to trick 5",
                'EAST plays tc',
                "EAST ready for SOUTH's card to trick 5",
                'EAST plays 6c',
                "EAST ready for SOUTH's card to trick 6",
                'WEST plays qc',
                "EAST ready for NORTH's card to trick 6",
                'WEST plays kc',
                "EAST ready for NORTH's card to trick 7",
                'EAST plays jc',
                "EAST ready for SOUTH's card to trick 7",
                'WEST plays 4d',
                "EAST ready for NORTH's card to trick 8",
                'EAST plays ad',
                "EAST ready for SOUTH's card to trick 8",
                'EAST plays 9c',
                "EAST ready for SOUTH's card to trick 9",
                'WEST plays 7h',
                "EAST ready for NORTH's card to trick 9",
                'WEST plays kh',
                "EAST ready for NORTH's card to trick 10",
                'EAST plays ah',
                "EAST ready for SOUTH's card to trick 10",
                'EAST plays ac',
                "EAST ready for SOUTH's card to trick 11",
                'WEST plays 8h',
                "EAST ready for NORTH's card to trick 11",
                'WEST plays qh',
                "EAST ready for NORTH's card to trick 12",
                'EAST plays 7d',
                "EAST ready for SOUTH's card to trick 12",
                'WEST plays jh',
                "EAST ready for NORTH's card to trick 13",
                'EAST plays qs',
            ]
        })
    bot = bluechip_bridge.BlueChipBridgeBot(game, 1, lambda: mock_client)
    state = game.new_initial_state()
    history = [
        33, 25, 3, 44, 47, 28, 23, 46, 1, 43, 30, 26, 29, 48, 24, 42, 13, 21,
        17, 8, 5, 34, 6, 7, 37, 49, 11, 38, 51, 32, 20, 9, 0, 14, 35, 22, 10,
        50, 15, 45, 39, 16, 12, 18, 27, 31, 41, 40, 4, 36, 19, 2, 52, 59, 52,
        61, 52, 62, 52, 68, 53, 70, 52, 71, 52, 74, 52, 76, 52, 82, 52, 52, 52,
        23, 7, 47, 31, 37, 25, 17, 45, 2, 10, 34, 30, 14, 6, 18, 4, 8, 0, 32,
        12, 16, 20, 40, 1, 44, 5, 36, 24, 9, 13, 49, 41, 28, 3, 22, 27, 46, 39,
        50, 11, 48, 15, 26, 29, 42, 33, 21, 19, 38, 51, 43, 35
    ]

    # Check the bot provides the expected actions
    for action in history:
      if state.current_player() == 1:
        bot_action = bot.step(state)
        self.assertEqual(action, bot_action)
      state.apply_action(action)

    # Check the session went as expected; send_line calls are us sending
    # data to the (mock) external bot.
    mock_client.assert_has_calls([
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('EAST ("WBridge5") seated'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line(
            'Teams: N/S "north-south" E/W "east-west"'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('start of board'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line(
            'Board number 1. Dealer NORTH. Neither vulnerable.'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line(
            "EAST's cards: C A J T 9 6. D A 8 7. H A T 5. S Q 9."),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH PASSES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH PASSES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('WEST bids 2D'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH PASSES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH PASSES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('WEST bids 3S'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH DOUBLES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH PASSES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('WEST bids 4D'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH PASSES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH PASSES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('WEST bids 5D'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH PASSES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH PASSES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('WEST PASSES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH PASSES'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays 7s'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line(
            "Dummy's cards: C K Q 4. D K 4. H K Q J 8 7 6 2. S 3."),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays ks'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays jd'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays 6d'),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('EAST to lead'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays 4h'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays 9h'),
        absltest.mock.call.send_line('EAST to lead'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays 3h'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays 3c'),
        absltest.mock.call.send_line('EAST to lead'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays 2c'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays 5c'),
        absltest.mock.call.send_line('EAST to lead'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays 7c'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays 2d'),
        absltest.mock.call.send_line('EAST to lead'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays 3d'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays 8c'),
        absltest.mock.call.send_line('EAST to lead'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays 5d'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays qd'),
        absltest.mock.call.send_line('EAST to lead'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays 2s'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays 8s'),
        absltest.mock.call.send_line('EAST to lead'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays js'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays 4s'),
        absltest.mock.call.send_line('EAST to lead'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays 5s'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays 9d'),
        absltest.mock.call.send_line('EAST to lead'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays td'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('SOUTH plays 6s'),
        absltest.mock.call.send_line('EAST to lead'),
        absltest.mock.call.read_line(),
        absltest.mock.call.read_line(),
        absltest.mock.call.send_line('NORTH plays as'),
        absltest.mock.call.read_line(),
    ])


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visualizing game trees with graphviz.

GameTree builds a `pygraphviz.AGraph` reprensentation of the game tree. The
resulting tree can be directly visualized in Jupyter notebooks or Google Colab
via SVG plotting - or written to a file by calling `draw(filename, prog="dot")`.

See `examples/treeviz_example.py` for a more detailed example.

This module relies on external dependencies, which need to be installed before
use. On a debian system follow these steps:
```
sudo apt-get install graphviz libgraphviz-dev
pip install pygraphviz
```
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import pyspiel

# pylint: disable=g-import-not-at-top
try:
  import pygraphviz
except (ImportError, Exception) as e:
  raise ImportError(
      str(e) + "\nPlease make sure to install the following dependencies:\n"
      "sudo apt-get install graphviz libgraphviz-dev\n"
      "pip install pygraphviz")
# pylint: enable=g-import-not-at-top

_PLAYER_SHAPES = {0: "square", 1: "ellipse"}
_PLAYER_COLORS = {-1: "black", 0: "blue", 1: "red"}
_FONTSIZE = 8
_WIDTH = _HEIGHT = 0.25
_ARROWSIZE = .5
_MARGIN = 0.01


def default_node_decorator(state):
  """Decorates a state-node of the game tree.

  This method can be called by a custom decorator to prepopulate the attributes
  dictionary. Then only relevant attributes need to be changed, or added.

  Args:
    state: The state.

  Returns:
    `dict` with graphviz node style attributes.
  """
  player = state.current_player()
  attrs = {
      "label": "",
      "fontsize": _FONTSIZE,
      "width": _WIDTH,
      "height": _HEIGHT,
      "margin": _MARGIN
  }
  if state.is_terminal():
    attrs["label"] = ", ".join(map(str, state.returns()))
    attrs["shape"] = "diamond"
  elif state.is_chance_node():
    attrs["shape"] = "point"
    attrs["width"] = _WIDTH / 2.
    attrs["height"] = _HEIGHT / 2.
  else:
    attrs["label"] = str(state.information_state_string())
    attrs["shape"] = _PLAYER_SHAPES.get(player, "ellipse")
    attrs["color"] = _PLAYER_COLORS.get(player, "black")
  return attrs


def default_edge_decorator(parent, unused_child, action):
  """Decorates a state-node of the game tree.

  This method can be called by a custom decorator to prepopulate the attributes
  dictionary. Then only relevant attributes need to be changed, or added.

  Args:
    parent: The parent state.
    unused_child: The child state, not used in the default decorator.
    action: `int` the selected action in the parent state.

  Returns:
    `dict` with graphviz node style attributes.
  """
  player = parent.current_player()
  attrs = {
      "label": " " + parent.action_to_string(player, action),
      "fontsize": _FONTSIZE,
      "arrowsize": _ARROWSIZE
  }
  attrs["color"] = _PLAYER_COLORS.get(player, "black")
  return attrs


class GameTree(pygraphviz.AGraph):
  """Builds `pygraphviz.AGraph` of the game tree.

  Attributes:
    game: A `pyspiel.Game` object.
    depth_limit: Maximum depth of the tree. Optional, default=-1 (no limit).
    node_decorator: Decorator function for nodes (states). Optional, default=
      `treeviz.default_node_decorator`.
    edge_decorator: Decorator function for edges (actions). Optional, default=
      `treeviz.default_edge_decorator`.
    group_terminal: Whether to display all terminal states at same level,
      default=False.
    group_infosets: Whether to group infosets together, default=False.
    group_pubsets: Whether to group public sets together, default=False.
    target_pubset: Whether to group all public sets "*" or a specific one.
    infoset_attrs: Attributes to style infoset grouping.
    pubset_attrs: Attributes to style public set grouping.
    kwargs: Keyword arguments passed on to `pygraphviz.AGraph.__init__`.
  """

  def __init__(self,
               game=None,
               depth_limit=-1,
               node_decorator=default_node_decorator,
               edge_decorator=default_edge_decorator,
               group_terminal=False,
               group_infosets=False,
               group_pubsets=False,
               target_pubset="*",
               infoset_attrs=None,
               pubset_attrs=None,
               **kwargs):

    kwargs["directed"] = kwargs.get("directed", True)
    super(GameTree, self).__init__(**kwargs)

    # We use pygraphviz.AGraph.add_subgraph to cluster nodes, and it requires a
    # default constructor. Thus game needs to be optional.
    if game is None:
      return

    self.game = game
    self._node_decorator = node_decorator
    self._edge_decorator = edge_decorator

    self._group_infosets = group_infosets
    self._group_pubsets = group_pubsets
    if self._group_infosets:
      if not self.game.get_type().provides_information_state_string:
        raise RuntimeError(
            "Grouping of infosets requested, but the game does not "
            "provide information state string.")
    if self._group_pubsets:
      if not self.game.get_type().provides_factored_observation_string:
        raise RuntimeError(
            "Grouping of public sets requested, but the game does not "
            "provide factored observations strings.")

    self._infosets = collections.defaultdict(lambda: [])
    self._pubsets = collections.defaultdict(lambda: [])
    self._terminal_nodes = []

    root = game.new_initial_state()
    self.add_node(self.state_to_str(root), **self._node_decorator(root))
    self._build_tree(root, 0, depth_limit)

    for (player, info_state), sibblings in self._infosets.items():
      cluster_name = "cluster_{}_{}".format(player, info_state)
      self.add_subgraph(sibblings, cluster_name,
                        **(infoset_attrs or {
                            "style": "dashed"
                        }))

    for pubset, sibblings in self._pubsets.items():
      if target_pubset == "*" or target_pubset == pubset:
        cluster_name = "cluster_{}".format(pubset)
        self.add_subgraph(sibblings, cluster_name,
                          **(pubset_attrs or {
                              "style": "dashed"
                          }))

    if group_terminal:
      self.add_subgraph(self._terminal_nodes, rank="same")

  def state_to_str(self, state):
    """Unique string representation of a state.

    Args:
      state: The state.

    Returns:
      String representation of state.
    """
    assert not state.is_simultaneous_node()
    # AGraph nodes can't have empty string == None as a key, thus we prepend " "
    return " " + state.history_str()

  def _build_tree(self, state, depth, depth_limit):
    """Recursively builds the game tree."""
    state_str = self.state_to_str(state)

    if state.is_terminal():
      self._terminal_nodes.append(state_str)
      return
    if depth > depth_limit >= 0:
      return

    for action in state.legal_actions():
      child = state.child(action)
      child_str = self.state_to_str(child)
      self.add_node(child_str, **self._node_decorator(child))
      self.add_edge(state_str, child_str,
                    **self._edge_decorator(state, child, action))

      if (self._group_infosets and not child.is_chance_node() and
          not child.is_terminal()):
        player = child.current_player()
        info_state = child.information_state_string()
        self._infosets[(player, info_state)].append(child_str)

      if self._group_pubsets:
        pub_obs_history = str(pyspiel.PublicObservationHistory(child))
        self._pubsets[pub_obs_history].append(child_str)

      self._build_tree(child, depth + 1, depth_limit)

  def _repr_svg_(self):
    """Allows to render directly in Jupyter notebooks and Google Colab."""
    if not self.has_layout:
      self.layout(prog="dot")
    return self.draw(format="svg").decode(self.encoding)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example for policy_aggregator_example.

Example.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags
import numpy as np

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import policy_aggregator

FLAGS = flags.FLAGS
flags.DEFINE_string("game_name", "kuhn_poker", "Game name")


class TestPolicy(policy.Policy):

  def __init__(self, action_int):
    self._action_int = action_int

  def action_probabilities(self, state, player_id=None):
    return {self._action_int: 1.0}


def main(unused_argv):
  env = rl_environment.Environment(FLAGS.game_name)

  policies = [[
      policy.TabularPolicy(env.game).copy_with_noise(alpha=float(i), beta=1.0)
      for i in range(2)
  ] for _ in range(2)]  # pylint: disable=g-complex-comprehension

  probabilities = [
      list(np.ones(len(policies[i])) / len(policies[i])) for i in range(2)
  ]

  pol_ag = policy_aggregator.PolicyAggregator(env.game)
  aggr_policies = pol_ag.aggregate([0, 1], policies, probabilities)

  exploitabilities = exploitability.nash_conv(env.game, aggr_policies)
  print("Exploitability : {}".format(exploitabilities))

  print(policies[0][0].action_probability_array)
  print(policies[0][1].action_probability_array)
  print(aggr_policies.policy)

  print("\nCopy Example")

  mother_policy = policy.TabularPolicy(env.game).copy_with_noise(1, 10)
  policies = [[mother_policy.__copy__() for _ in range(2)] for _ in range(2)]
  probabilities = [
      list(np.ones(len(policies)) / len(policies)) for _ in range(2)
  ]

  pol_ag = policy_aggregator.PolicyAggregator(env.game)
  aggr_policy = pol_ag.aggregate([0], policies, probabilities)

  for state, value in aggr_policy.policy[0].items():
    polici = mother_policy.policy_for_key(state)

    value_normal = {
        action: probability
        for action, probability in enumerate(polici)
        if probability > 0
    }
    for key in value.keys():
      print("State : {}. Key : {}. Aggregated : {}. Real : {}. Passed : {}"
            .format(state, key, value[key], value_normal[key],
                    np.abs(value[key] - value_normal[key]) < 1e-8))


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Computes a softmax policy from a value function."""
from typing import Optional

import numpy as np

from open_spiel.python import policy
from open_spiel.python.mfg import value


class SoftmaxPolicy(policy.Policy):
  """Computes the softmax policy of a value function."""

  def __init__(self,
               game,
               player_ids,
               temperature: float,
               state_action_value: value.ValueFunction,
               prior_policy: Optional[policy.Policy] = None):
    """Initializes the softmax policy.

    Args:
      game: The game to analyze.
      player_ids: list of player ids for which this policy applies; each
        should be in the range 0..game.num_players()-1.
      temperature: float to scale the values (multiplied by 1/temperature).
      state_action_value: A state-action value function.
      prior_policy: Optional argument. Prior policy to scale the softmax
        policy.
    """
    super(SoftmaxPolicy, self).__init__(game, player_ids)
    self._state_action_value = state_action_value
    self._prior_policy = prior_policy
    self._temperature = temperature

  def action_probabilities(self, state, player_id=None):
    legal_actions = state.legal_actions()
    max_q = np.max(
        [self._state_action_value(state, action) for action in legal_actions])
    exp_q = [
        np.exp((self._state_action_value(state, action) - max_q) /
               self._temperature) for action in legal_actions
    ]
    if self._prior_policy is not None:
      prior_probs = self._prior_policy.action_probabilities(state)
      exp_q = [
          prior_probs.get(action, 0) * exp_q[i]
          for i, action in enumerate(legal_actions)
      ]
    denom = sum(exp_q)
    smax_q = exp_q if denom == 0 else exp_q / denom
    return dict(zip(legal_actions, smax_q))



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A class to log stuff to a file, mainly useful in parallel situations."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import os


class FileLogger(object):
  """A logger to print stuff to a file."""

  def __init__(self, path, name, quiet=False, also_to_stdout=False):
    self._fd = open(os.path.join(path, "log-{}.txt".format(name)), "w")
    self._quiet = quiet
    self.also_to_stdout = also_to_stdout

  def print(self, *args):
    # Date/time with millisecond precision.
    date_prefix = "[{}]".format(datetime.datetime.now().isoformat(" ")[:-3])
    print(date_prefix, *args, file=self._fd, flush=True)
    if self.also_to_stdout:
      print(date_prefix, *args, flush=True)

  def opt_print(self, *args):
    if not self._quiet:
      self.print(*args)

  def __enter__(self):
    return self

  def __exit__(self, unused_exception_type, unused_exc_value, unused_traceback):
    self.close()

  def close(self):
    if self._fd:
      self._fd.close()
      self._fd = None

  def __del__(self):
    self.close()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.cfr."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import cfr
from open_spiel.python.algorithms import expected_game_score
from open_spiel.python.algorithms import exploitability
import pyspiel

_KUHN_GAME = pyspiel.load_game("kuhn_poker")
_LEDUC_GAME = pyspiel.load_game("leduc_poker")

_KUHN_UNIFORM_POLICY = policy.TabularPolicy(_KUHN_GAME)
_LEDUC_UNIFORM_POLICY = policy.TabularPolicy(_LEDUC_GAME)


class ModuleLevelFunctionTest(absltest.TestCase):

  def test__update_current_policy(self):
    game = pyspiel.load_game("kuhn_poker")
    tabular_policy = policy.TabularPolicy(game)

    cumulative_regrets = np.arange(0, 12 * 2).reshape((12, 2))
    expected_policy = cumulative_regrets / np.sum(
        cumulative_regrets, axis=-1, keepdims=True)
    nodes_indices = {
        u"0": 0,
        u"0pb": 1,
        u"1": 2,
        u"1pb": 3,
        u"2": 4,
        u"2pb": 5,
        u"1p": 6,
        u"1b": 7,
        u"2p": 8,
        u"2b": 9,
        u"0p": 10,
        u"0b": 11,
    }
    # pylint: disable=g-complex-comprehension
    info_state_nodes = {
        key: cfr._InfoStateNode(
            legal_actions=[0, 1],
            index_in_tabular_policy=None,
            cumulative_regret=dict(enumerate(cumulative_regrets[index])),
            cumulative_policy=None) for key, index in nodes_indices.items()
    }
    # pylint: enable=g-complex-comprehension

    cfr._update_current_policy(tabular_policy, info_state_nodes)

    np.testing.assert_array_equal(expected_policy,
                                  tabular_policy.action_probability_array)


class CFRTest(parameterized.TestCase, absltest.TestCase):

  @parameterized.parameters(
      list(itertools.product([True, False], [True, False], [True, False])))
  def test_policy_zero_is_uniform(self, linear_averaging, regret_matching_plus,
                                  alternating_updates):
    # We use Leduc and not Kuhn, because Leduc has illegal actions and Kuhn does
    # not.
    game = pyspiel.load_game("leduc_poker")
    cfr_solver = cfr._CFRSolver(
        game,
        regret_matching_plus=regret_matching_plus,
        linear_averaging=linear_averaging,
        alternating_updates=alternating_updates)

    np.testing.assert_array_equal(
        _LEDUC_UNIFORM_POLICY.action_probability_array,
        cfr_solver.current_policy().action_probability_array)
    np.testing.assert_array_equal(
        _LEDUC_UNIFORM_POLICY.action_probability_array,
        cfr_solver.average_policy().action_probability_array)

  def test_cfr_kuhn_poker(self):
    game = pyspiel.load_game("kuhn_poker")
    cfr_solver = cfr.CFRSolver(game)
    for _ in range(300):
      cfr_solver.evaluate_and_update_policy()
    average_policy = cfr_solver.average_policy()
    average_policy_values = expected_game_score.policy_value(
        game.new_initial_state(), [average_policy] * 2)
    # 1/18 is the Nash value. See https://en.wikipedia.org/wiki/Kuhn_poker
    np.testing.assert_allclose(
        average_policy_values, [-1 / 18, 1 / 18], atol=1e-3)

  def test_cfr_plus_kuhn_poker(self):
    game = pyspiel.load_game("kuhn_poker")
    cfr_solver = cfr.CFRPlusSolver(game)
    for _ in range(200):
      cfr_solver.evaluate_and_update_policy()
    average_policy = cfr_solver.average_policy()
    average_policy_values = expected_game_score.policy_value(
        game.new_initial_state(), [average_policy] * 2)
    # 1/18 is the Nash value. See https://en.wikipedia.org/wiki/Kuhn_poker
    np.testing.assert_allclose(
        average_policy_values, [-1 / 18, 1 / 18], atol=1e-3)

  def test_cfr_plus_solver_best_response_mdp(self):
    game = pyspiel.load_game("kuhn_poker")
    cfr_solver = cfr.CFRPlusSolver(game)
    for _ in range(200):
      cfr_solver.evaluate_and_update_policy()
    average_policy = cfr_solver.average_policy()
    pyspiel_avg_policy = policy.python_policy_to_pyspiel_policy(average_policy)
    br_computer = pyspiel.TabularBestResponseMDP(game, pyspiel_avg_policy)
    br_info = br_computer.exploitability()
    self.assertLessEqual(br_info.exploitability, 0.001)

  def test_cfr_cce_ce_dist_goofspiel(self):
    """Copy of the TestCCEDistCFRGoofSpiel in corr_dist_test.cc."""
    game = pyspiel.load_game(
        "turn_based_simultaneous_game(game=goofspiel(num_cards=3,points_order="
        "descending,returns_type=total_points))")
    for num_iterations in [1, 10, 100]:
      policies = []
      cfr_solver = cfr.CFRSolver(game)
      for _ in range(num_iterations):
        cfr_solver.evaluate_and_update_policy()
        policies.append(
            policy.python_policy_to_pyspiel_policy(cfr_solver.current_policy()))
      mu = pyspiel.uniform_correlation_device(policies)
      cce_dist_info = pyspiel.cce_dist(game, mu)
      print("goofspiel, cce test num_iters: {}, cce_dist: {}, per player: {}"
            .format(num_iterations, cce_dist_info.dist_value,
                    cce_dist_info.deviation_incentives))
      # Try converting one of the BR policies:
      _ = policy.pyspiel_policy_to_python_policy(
          game, cce_dist_info.best_response_policies[0])

      # Assemble the same correlation device manually, just as an example for
      # how to do non-uniform distributions of them and to test the python
      # bindings for lists of tuples works properly
      uniform_prob = 1.0 / len(policies)
      mu2 = [(uniform_prob, policy) for policy in policies]
      cce_dist_info2 = pyspiel.cce_dist(game, mu2)
      self.assertAlmostEqual(cce_dist_info2.dist_value,
                             sum(cce_dist_info.deviation_incentives))
      # Test the CEDist function too, why not. Disable the exact one, as it
      # takes too long for a test.
      # ce_dist_info = pyspiel.ce_dist(game, pyspiel.determinize_corr_dev(mu))
      ce_dist_info = pyspiel.ce_dist(
          game, pyspiel.sampled_determinize_corr_dev(mu, 100))
      print("goofspiel, ce test num_iters: {}, ce_dist: {}, per player: {}"
            .format(num_iterations, ce_dist_info.dist_value,
                    ce_dist_info.deviation_incentives))
      print("number of conditional best responses per player:")
      for p in range(game.num_players()):
        print("  player {}, num: {}".format(
            p, len(ce_dist_info.conditional_best_response_policies[p])))

  @parameterized.parameters(
      list(itertools.product([True, False], [True, False], [True, False])))
  def test_cfr_kuhn_poker_runs_with_multiple_players(self, linear_averaging,
                                                     regret_matching_plus,
                                                     alternating_updates):
    num_players = 3

    game = pyspiel.load_game("kuhn_poker", {"players": num_players})
    cfr_solver = cfr._CFRSolver(
        game,
        regret_matching_plus=regret_matching_plus,
        linear_averaging=linear_averaging,
        alternating_updates=alternating_updates)
    for _ in range(10):
      cfr_solver.evaluate_and_update_policy()
    average_policy = cfr_solver.average_policy()
    average_policy_values = expected_game_score.policy_value(
        game.new_initial_state(), [average_policy] * num_players)
    del average_policy_values

  @parameterized.parameters(list(itertools.product([False, True])))
  def test_simultaneous_two_step_avg_1b_seq_in_kuhn_poker(
      self, regret_matching_plus):
    num_players = 2
    game = pyspiel.load_game("kuhn_poker", {"players": num_players})
    cfr_solver = cfr._CFRSolver(
        game,
        regret_matching_plus=regret_matching_plus,
        linear_averaging=False,
        alternating_updates=False)

    def check_avg_policy_is_uniform_random():
      avg_policy = cfr_solver.average_policy()
      for player_info_states in avg_policy.states_per_player:
        for info_state in player_info_states:
          state_policy = avg_policy.policy_for_key(info_state)
          np.testing.assert_allclose(state_policy, [1.0 / len(state_policy)] *
                                     len(state_policy))

    check_avg_policy_is_uniform_random()

    cfr_solver.evaluate_and_update_policy()
    check_avg_policy_is_uniform_random()

    cfr_solver.evaluate_and_update_policy()

    # The acting player in 1b is player 1 and they have not acted before, so
    # the probability this player plays to this information state is 1, and
    # the sequence probability of any action is just the probability of that
    # action given the information state. On the first iteration, this
    # probability is 0.5 for both actions. On the second iteration, the
    # current policy is [0, 1], so the average cumulants should be
    # [0.5, 1.5]. Normalizing this gives the average policy.
    normalization = 0.5 + 0.5 + 1
    np.testing.assert_allclose(cfr_solver.average_policy().policy_for_key("1b"),
                               [0.5 / normalization, (0.5 + 1) / normalization])

  def test_policy(self):
    game = pyspiel.load_game("kuhn_poker")
    solver = cfr.CFRPlusSolver(game)

    tabular_policy = solver.current_policy()
    self.assertLen(tabular_policy.state_lookup, 12)
    for info_state_str in tabular_policy.state_lookup.keys():
      np.testing.assert_equal(
          np.asarray([0.5, 0.5]), tabular_policy.policy_for_key(info_state_str))

  @parameterized.parameters([
      (pyspiel.load_game("kuhn_poker"), pyspiel.CFRSolver, cfr.CFRSolver),
      (pyspiel.load_game("leduc_poker"), pyspiel.CFRSolver, cfr.CFRSolver),
      (pyspiel.load_game("kuhn_poker"), pyspiel.CFRPlusSolver,
       cfr.CFRPlusSolver),
      (pyspiel.load_game("leduc_poker"), pyspiel.CFRPlusSolver,
       cfr.CFRPlusSolver),
  ])
  def test_cpp_algorithms_identical_to_python_algorithm(self, game, cpp_class,
                                                        python_class):
    cpp_solver = cpp_class(game)
    python_solver = python_class(game)

    for _ in range(5):
      cpp_solver.evaluate_and_update_policy()
      python_solver.evaluate_and_update_policy()

      cpp_avg_policy = cpp_solver.average_policy()
      python_avg_policy = python_solver.average_policy()

      # We do not compare the policy directly as we do not have an easy way to
      # convert one to the other, so we use the exploitability as a proxy.
      cpp_expl = pyspiel.nash_conv(game, cpp_avg_policy)
      python_expl = exploitability.nash_conv(game, python_avg_policy)
      self.assertEqual(cpp_expl, python_expl)
    # Then we also check the CurrentPolicy, just to check it is giving the same
    # results too
    cpp_current_policy = cpp_solver.current_policy()
    python_current_policy = python_solver.current_policy()
    cpp_expl = pyspiel.nash_conv(game, cpp_current_policy)
    python_expl = exploitability.nash_conv(game, python_current_policy)
    self.assertEqual(cpp_expl, python_expl)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reservoir buffer implemented in Numpy.

See https://en.wikipedia.org/wiki/Reservoir_sampling for more details.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import random
import numpy as np


# TODO(perolat): refactor the reservoir with the NFSP Pytorch implementation
class ReservoirBuffer(object):
  """Allows uniform sampling over a stream of data.

  This class supports the storage of arbitrary elements, such as observation
  tensors, integer actions, etc.

  See https://en.wikipedia.org/wiki/Reservoir_sampling for more details.
  """

  def __init__(self, reservoir_buffer_capacity):
    self._reservoir_buffer_capacity = reservoir_buffer_capacity
    self._data = []
    self._add_calls = 0

  def add(self, element):
    """Potentially adds `element` to the reservoir buffer.

    Args:
      element: data to be added to the reservoir buffer.
    """
    if len(self._data) < self._reservoir_buffer_capacity:
      self._data.append(element)
    else:
      idx = np.random.randint(0, self._add_calls + 1)
      if idx < self._reservoir_buffer_capacity:
        self._data[idx] = element
    self._add_calls += 1

  def sample(self, num_samples):
    """Returns `num_samples` uniformly sampled from the buffer.

    Args:
      num_samples: `int`, number of samples to draw.

    Returns:
      An iterable over `num_samples` random elements of the buffer.

    Raises:
      ValueError: If there are less than `num_samples` elements in the buffer
    """
    if len(self._data) < num_samples:
      raise ValueError("{} elements could not be sampled from size {}".format(
          num_samples, len(self._data)))
    return random.sample(self._data, num_samples)

  def clear(self):
    self._data = []
    self._add_calls = 0

  def __len__(self):
    return len(self._data)

  def __iter__(self):
    return iter(self._data)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python spiel example."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags
import numpy as np

from open_spiel.python.bots import human
from open_spiel.python.bots import uniform_random
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_integer("seed", 12761381, "The seed to use for the RNG.")

# Supported types of players: "random", "human", "check_call", "fold"
flags.DEFINE_string("player0", "random", "Type of the agent for player 0.")
flags.DEFINE_string("player1", "random", "Type of the agent for player 1.")


def LoadAgent(agent_type, game, player_id, rng):
  """Return a bot based on the agent type."""
  if agent_type == "random":
    return uniform_random.UniformRandomBot(player_id, rng)
  elif agent_type == "human":
    return human.HumanBot()
  elif agent_type == "check_call":
    policy = pyspiel.PreferredActionPolicy([1, 0])
    return pyspiel.make_policy_bot(game, player_id, FLAGS.seed, policy)
  elif agent_type == "fold":
    policy = pyspiel.PreferredActionPolicy([0, 1])
    return pyspiel.make_policy_bot(game, player_id, FLAGS.seed, policy)
  else:
    raise RuntimeError("Unrecognized agent type: {}".format(agent_type))


def main(_):
  rng = np.random.RandomState(FLAGS.seed)

  # Make sure poker is compiled into the library, as it requires an optional
  # dependency: the ACPC poker code. To ensure it is compiled in, prepend both
  # the install.sh and build commands with OPEN_SPIEL_BUILD_WITH_ACPC=ON.
  # See here:
  # https://github.com/deepmind/open_spiel/blob/master/docs/install.md#configuration-conditional-dependencies
  # for more details on optional dependencies.
  games_list = pyspiel.registered_names()
  assert "universal_poker" in games_list

  fcpa_game_string = pyspiel.hunl_game_string("fcpa")
  print("Creating game: {}".format(fcpa_game_string))
  game = pyspiel.load_game(fcpa_game_string)

  agents = [
      LoadAgent(FLAGS.player0, game, 0, rng),
      LoadAgent(FLAGS.player1, game, 1, rng)
  ]

  state = game.new_initial_state()

  # Print the initial state
  print("INITIAL STATE")
  print(str(state))
  
 
  while not state.is_terminal():
    # The state can be three different types: chance node,
    # simultaneous node, or decision node
    current_player = state.current_player()
    if state.is_chance_node():
      # Chance node: sample an outcome
      outcomes = state.chance_outcomes()
      num_actions = len(outcomes)
      print("Chance node with " + str(num_actions) + " outcomes")
      action_list, prob_list = zip(*outcomes)
      action = rng.choice(action_list, p=prob_list)
      print("Sampled outcome: ",
            state.action_to_string(state.current_player(), action))
      state.apply_action(action)
    else:
      # Decision node: sample action for the single current player
      legal_actions = state.legal_actions()
      for action in legal_actions:
        print("Legal action: {} ({})".format(
            state.action_to_string(current_player, action), action))
      action = agents[current_player].step(state)
      action_string = state.action_to_string(current_player, action)
      print("Player ", current_player, ", chose action: ",
            action_string)
      state.apply_action(action)

    print("")
    print("NEXT STATE:")
    print(str(state))


  # This is a minor test
  print("\n\nRESETTING HOLE CARDS BEFORE RETURNS CALCULATED\n\n")
  partialState = [[0, 4], [50, 51]]
  state.set_partial_game_state(partialState)
  print(str(state))
  print(str(state.hole_cards(0)))

  # Game is now done. Print utilities for each player
  returns = state.returns()
  for pid in range(game.num_players()):
    print("Utility for player {} is {}".format(pid, returns[pid]))


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the C++ matrix game utility methods exposed to Python."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest

from open_spiel.python.algorithms import lp_solver
import pyspiel


class MatrixGamesUtilsTest(absltest.TestCase):

  def test_num_deterministic_policies(self):
    # Kuhn poker has six information sets with two actions each (2^6 = 64).
    game = pyspiel.load_game("kuhn_poker")
    self.assertEqual(pyspiel.num_deterministic_policies(game, 0), 64)
    self.assertEqual(pyspiel.num_deterministic_policies(game, 1), 64)
    # Leduc poker has larger than 2^64 - 1, so -1 will be returned.
    game = pyspiel.load_game("leduc_poker")
    self.assertEqual(pyspiel.num_deterministic_policies(game, 0), -1)
    self.assertEqual(pyspiel.num_deterministic_policies(game, 1), -1)

  def test_extensive_to_matrix_game(self):
    kuhn_game = pyspiel.load_game("kuhn_poker")
    kuhn_matrix_game = pyspiel.extensive_to_matrix_game(kuhn_game)
    unused_p0_strategy, unused_p1_strategy, p0_sol_val, p1_sol_val = (
        lp_solver.solve_zero_sum_matrix_game(kuhn_matrix_game))
    # value from Kuhn 1950 or https://en.wikipedia.org/wiki/Kuhn_poker
    self.assertAlmostEqual(p0_sol_val, -1 / 18)
    self.assertAlmostEqual(p1_sol_val, +1 / 18)

  def test_extensive_to_matrix_game_type(self):
    game = pyspiel.extensive_to_matrix_game(pyspiel.load_game("kuhn_poker"))
    game_type = game.get_type()
    self.assertEqual(game_type.dynamics, pyspiel.GameType.Dynamics.SIMULTANEOUS)
    self.assertEqual(game_type.chance_mode,
                     pyspiel.GameType.ChanceMode.DETERMINISTIC)
    self.assertEqual(game_type.information,
                     pyspiel.GameType.Information.ONE_SHOT)
    self.assertEqual(game_type.utility, pyspiel.GameType.Utility.ZERO_SUM)

  def test_extensive_to_matrix_game_payoff_matrix(self):
    turn_based_game = pyspiel.load_game_as_turn_based("matrix_pd")
    matrix_game = pyspiel.extensive_to_matrix_game(turn_based_game)
    orig_game = pyspiel.load_matrix_game("matrix_pd")

    for row in range(orig_game.num_rows()):
      for col in range(orig_game.num_cols()):
        for player in range(2):
          self.assertEqual(
              orig_game.player_utility(player, row, col),
              matrix_game.player_utility(player, row, col))


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""GAMUT games.

See https://github.com/deepmind/open_spiel/tree/master/open_spiel/games/gamut
for details on how to build OpenSpiel with support for GAMUT.
"""

from absl import logging  # pylint:disable=unused-import

import numpy as np

from open_spiel.python.egt.utils import game_payoffs_array
import pyspiel


class GAMUT(object):
  """GAMUT Games."""

  def __init__(self, config_list, java_path='', seed=None):
    """Ctor. Inits payoff tensor (players x actions x ... np.array).

    Args:
      config_list: a list or strings alternating between gamut flags and values
        see http://gamut.stanford.edu/userdoc.pdf for more information
        e.g., config_list = ['-g', 'CovariantGame', '-players', '6',
                             '-normalize', '-min_payoff', '0',
                             '-max_payoff', '1', '-actions', '5', '-r', '0']
      java_path: string, java path
      seed: random seed, some GAMUT games are randomly generated
    """
    self.pt = None
    self.config_list = config_list

    self.seed = seed
    self.random = np.random.RandomState(seed)

    # parse interval for rho if supplied, e.g., '[-.2,1]'
    if '-r' in config_list:
      idx = next(i for i, s in enumerate(config_list) if s == '-r')
      val = config_list[idx + 1]
      if not val.isnumeric() and val[0] in '([' and val[-1] in ')]':
        a, b = val.strip('[]()').split(',')
        a = float(a)
        b = float(b)
        rho = self.random.rand() * (b - a) + a
        config_list[idx + 1] = str(rho)

    if isinstance(seed, int):
      self.config_list += ['-random_seed', str(seed)]
    self.java_path = java_path

    if java_path:
      generator = pyspiel.GamutGenerator(
          java_path,
          'gamut/gamut_main_deploy.jar')
    else:  # use default java path as specified by pyspiel
      generator = pyspiel.GamutGenerator(
          'gamut.jar')
    self.game = generator.generate_game(config_list)

  def num_players(self):
    return self.game.num_players()

  def num_strategies(self):
    return [self.game.num_distinct_actions()] * self.num_players()

  def payoff_tensor(self):
    if self.pt is None:
      pt = np.asarray(game_payoffs_array(self.game))
      self.pt = pt - self.game.min_utility()
    return self.pt

  def get_payoffs_for_strategies(self, policies):
    """Return vector of payoffs for all players given list of strategies.

    Args:
      policies: list of integers indexing strategies for each player
    Returns:
      np.array (length num players) of payoffs
    """
    state = self.game.new_initial_state()
    state.apply_actions(policies)
    return np.asarray(state.returns()) - self.game.min_utility()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Useful functions for testing."""

from typing import Optional
import numpy as np
import pyspiel


def random_playout(state: pyspiel.State, seed: Optional[int] = None):
  """Plays random actions until the state is terminal."""
  rng = np.random.RandomState(seed)
  while not state.is_terminal():
    state.apply_action(rng.choice(state.legal_actions()))
  return state
